
import pandas as pd
import numpy as np
from scipy import sparse
from sklearn.cluster import AffinityPropagation as AP

import tensorflow as tf
import lightgbm as lgb
from six.moves import cPickle
import os
import time
import re
import subprocess as sp
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.svm import LinearSVC, SVC
from sklearn.preprocessing import label_binarize
from sklearn.metrics import roc_auc_score
from sklearn.kernel_approximation import RBFSampler


class ClusterPairwise(object):

    # 使用hhblits检索fasFilePath, 序列检索自身, 得到的结果存放在hhr_csvFormat_path
    # 该函数主要的目的是从hhr_csvFormat_path对应的文件中得出protein similarity network
    # 参数hhr_df表示的就是hhblits检索结果构成的dataframe(pandas中的)
    # query template prob evalue score是必然存在的列
    # spec_df记录的是序列在matrix（特征矩阵）中的位置（第几行）
    # 这里使用它的目的是方便用数字代表序列
    # 结果返回一个稀疏矩阵sparse.lil_matrix类型
    def hhr2network(self, hhr_df, spec_df):
        # query 和 template统称为template，只不过用途不一样，一个是用于检索，一个是被检索
        hhr_df.loc[:,["query","template"]] = \
            hhr_df.loc[:,["query","template"]].apply(
                lambda x: \
                    pd.Series(spec_df.loc[x,"row num"].reset_index(drop=True),
                        name=x.name)
            )
        
        # pairwise_df = hhr_df.loc[:,["query","template","prob"]]

        r,_ = hhr_df.shape
        z = np.zeros((r,r),dtype=np.float32)
        sim = sparse.lil_matrix(z)

        sim[hhr_df.loc[:,"query"],hhr_df.loc[:,"template"]] = hhr_df.loc[:,"prob"]

        # 对称化
        sim = sim.maximum(sim.transpose())

        # 去掉自环
        sim[range(r),range(r)] = 0.0

        return sim
    
    # sim是一个相似度矩阵，表示pairwise的similarity值
    # 根据sim来得到聚类结果
    def affinity_propagation(self, sim):
        random_state = np.random.seed()
        model = AP(damping=0.73,affinity='precomputed',random_state=random_state)

        clustering = model.fit(sim)

        return clustering

    def expand_template(self, t_df, kv_df):
        m_df = t_df.join(kv_df.set_index("template"),on="template")
        return m_df
    
    def renumber(self, m_df):
        N = m_df.shape[0]
        m_df = m_df.astype({"row num":int})
        m_df = m_df.sort_values(by="row num")
        # 重新编号
        m_df = m_df.assign(new_row_num=list(range(N)))
        return m_df

    # 计算序列与superfamily type对应的打分结果
    # m_df包含template, sf, row num, new_row_num这几列
    # R是排序结果，tensor类型，大小为TensorShape([n_samples])
    def S2C_transform(self, R, m_df):
        idx_list = []
        for _, g in m_df.groupby("sf"):
            idx = g.loc[:,"new_row_num"].to_numpy(int)
            idx = idx.reshape((-1,1))
            idx_list.append(idx)

        update_list = []
        for idx in idx_list:
            p0 = tf.gather_nd(R, idx)
            p1 = tf.math.reduce_mean(1 / (1 + tf.math.log(1/p0)))
            u = p1 * tf.ones(idx.shape[0])
            update_list.append(u)
        
        update_con = tf.concat(update_list,axis=0)

        indices = np.vstack(idx_list)
        # print(indices.shape)
        updates = tf.reshape(update_con,(-1,))
        # print(updates.shape)

        R = tf.tensor_scatter_nd_update(R, indices,updates)
        return R


# 尝试通过LightGBM的Dataset构造数据集，同时进行增量训练
# 因为构建Dataset，有个free_raw_data，可以保证在构造Dataset后，释放原始数据，这保证了内存占用不会太大
class Ranker(object):
    # 根据所有的检索结果生成pairwise
    # srched_list是所有srch_df组成的列表
    # srch_df表示检索结果，是一个包含query, template的dataframe
    # vsrch_df格式与srch_df一样表示的是validation数据集检索training samples的结果
    def gen_pariwise(self, cv_df, srch_df, vsrch_df):
        for fold_num in cv_df.columns:
            idx_df = srch_df.apply(lambda x: x.isin(cv_df.loc[:,fold_num]))
            pairwise_df = srch_df.loc[~(idx_df.loc[:,"query"] == idx_df.loc[:,"template"])]
            vpairwise_df = vsrch_df.loc[~vsrch_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
            yield pairwise_df, vpairwise_df

    def similarity(self, **kArgs):
        assert kArgs.get("rep") is not None
        ron = tf.constant(kArgs["rep"],tf.float32)

        if kArgs.get("queryName") is not None and\
            kArgs.get("template_ser") is not None:
            tidx = kArgs["template_ser"].to_numpy().reshape((-1,1))
        
            dis = ron[kArgs["queryName"]] - tf.gather_nd(ron, tidx)
        elif kArgs.get("pairwise_df") is not None:
            pairwise_df = kArgs["pairwise_df"]
            lidx = pairwise_df.loc[:, "query"].to_numpy().reshape((-1,1))
            ridx = pairwise_df.loc[:, "template"].to_numpy().reshape((-1,1))
            dis = tf.gather_nd(ron, lidx) - tf.gather_nd(ron, ridx)
        elif kArgs.get("vpairwise_df") is not None and\
            kArgs.get("vrep") is not None:

            # print(kArgs)
            lidx = kArgs["vpairwise_df"].loc[:,"query"].to_numpy().reshape((-1,1))
            ridx = kArgs["vpairwise_df"].loc[:,"template"].to_numpy().reshape((-1,1))
            vron = tf.constant(kArgs["vrep"],tf.float32)
            
            # print(tf.gather_nd(vron, lidx))
            # print(tf.gather_nd(ron, ridx))
            dis = tf.gather_nd(vron, lidx) - tf.gather_nd(ron, ridx)
        else:
            raise AttributeError("可支持的参数，queryName, template_ser和rep或vpairwise_df,rep和vrep")
        
        dis = tf.abs(dis)
        maximum = tf.linalg.norm(dis, ord=np.inf,axis=1,keepdims=True)
        log = tf.math.log(
            tf.linalg.norm(
                dis,ord=2,axis=1,keepdims=True
            )
        )
        log /= tf.math.log(3.)
        sim = tf.concat([dis, maximum, log],axis=1)
            
        return sim
    
    
    # cv_prepare(self, queryName,train_df,pairwise_df, vpairwise_df,spec_df, vspec_df, kv_dict, kv_df):
    # train_df是query name为index，"行下标"为列的datagrame，表示trainin samples在特征矩阵中的位置
    # pairwise_df是包含query template这两列的dataframe，表示的是training samples构成的pairwise
    # vpairwise_df与pairwise_df类似，表示的是validation数据集与training samples构成的pairwise
    # spec_df与train_df一样
    # vspec_df与train_df格式一样，但表示的是validation数据集中的序列在特征矩阵中的位置
    # kv_dict和kv_df表示的是序列对应的超家族
    # 分别是字典(query name为key, superfamily为value)，dataframe格式(query name为index，superfamily为列)
    def cv_prepare(self, **kArgs):
        result = {} # 用于返回结果的字典类型文件

        assert (kArgs.get("kv_dict") is not None and \
            kArgs.get("kv_df") is not None)

        kv_dict = kArgs["kv_dict"]
        kv_df = kArgs["kv_df"]

        assert (kArgs.get("spec_df") is not None and\
            kArgs.get("vspec_df") is not None)
        spec_df = kArgs["spec_df"]
        vspec_df = kArgs["vspec_df"]

        if kArgs.get("queryName") is not None and \
            kArgs.get("train_df") is not None:
            queryName = kArgs.get("queryName")
            train_df = kArgs["train_df"]
            q_sf = kv_dict[queryName]
            label = (train_df.join(kv_df).loc[:,"sf"] == q_sf).to_numpy(np.int64)
            queryName = spec_df.loc[queryName,"row num"]
            result["queryName"] = queryName
            result["label"] = label
        elif kArgs.get("pairwise_df") is not None:
            pairwise_df = kArgs["pairwise_df"]
            pairwise_df = pairwise_df.join(kv_df,on="query")
            pairwise_df = pairwise_df.join(kv_df,lsuffix="_query",rsuffix="_template",on="template")
            label = (pairwise_df.loc[:, "sf_query"] == pairwise_df.loc[:,"sf_template"]).to_numpy(np.int64)

            pairwise_df = pairwise_df.join(vspec_df,on="query")
            pairwise_df = pairwise_df.join(spec_df,lsuffix="_query",rsuffix="_template",on="template")
            # print(pd.isna(pairwise_df).any())
            pairwise_df = pairwise_df.loc[:,["row num_query","row num_template"]]
            pairwise_df = pairwise_df.rename(columns={"row num_query":"query","row num_template":"template"})           
            result["label"] = label
            result["pairwise_df"] = pairwise_df
        else:
            raise AssertionError("queryName和train_df或pairwise_df必须提供一个")

        assert kArgs.get("vpairwise_df") is not None

        vpairwise_df = kArgs["vpairwise_df"]
        vpairwise_df = vpairwise_df.join(kv_df,on="query")
        vpairwise_df = vpairwise_df.join(kv_df,lsuffix="_query",rsuffix="_template",on="template")
        # print(pd.isna(vpairwise_df).any())
        vlabel = (vpairwise_df.loc[:, "sf_query"] == vpairwise_df.loc[:,"sf_template"]).to_numpy(np.int64)

        vpairwise_df = vpairwise_df.join(vspec_df,on="query")
        vpairwise_df = vpairwise_df.join(spec_df,lsuffix="_query",rsuffix="_template",on="template")
        # print(pd.isna(vpairwise_df).any())
        vpairwise_df = vpairwise_df.loc[:,["row num_query","row num_template"]]
        vpairwise_df = vpairwise_df.rename(columns={"row num_query":"query","row num_template":"template"})
        # print(pd.isna(vpairwise_df).any())

        result["vlabel"] = vlabel
        result["vpairwise_df"] = vpairwise_df

        return result

    # X, y, eX, ey都是numpy.ndarray的类型
    def cv_training(self, X,y, eX, ey,params,group=None,egroup=None,init_model=None):
        lgb_train = lgb.Dataset(
            X,y,group=group,free_raw_data=True
        )
        lgb_val = lgb.Dataset(
            eX,ey,group=egroup,free_raw_data=True
        )

        print(params)
        trained = lgb.train(
            params,
            lgb_train,
            valid_sets=[lgb_val],
            init_model=init_model,
            early_stopping_rounds=100,
            verbose_eval=10
        )

        # return lgb_train, lgb_val # 用于测试

        return trained

    # cv_df表示cross_validation的划分情况，一列表示一折
    # 在kArgs中
    # 提供了template_df则根据数据集划分情况生成对应用于训练的template_df
    # 提供了vtemplate_df则表示这是pairwise类型的检索结果，同样根据数据集划分情况进行生成
    # 体统了query_df则表示这是测试集，根据数据划分情况进行生成，注意区分query生成方法与其他两种的不同之处
    def training_generator(self, cv_df, **kArgs):
        assert kArgs.get("template_df") is not None

        template_df = kArgs["template_df"]
        result = {}

        assert "template" in template_df.columns

        for fold_num in cv_df.columns:
            # 如果存在query
            # 就表示这是pairwise类型的检索结果
            # 否则就表示这是只有template(或者还有row num)的dataframe
            if "query" in template_df.columns:
                template_df = template_df.loc[~template_df.loc[:,"query"].isin(cv_df.loc[:,fold_num])]
            template_df = template_df.loc[~template_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
            result["template"] = template_df

            # validaiont数据集如果有query也不用处理，因为validatoin数据集和training数据集是不重叠的
            vtemplate_df = kArgs["vtemplate_df"]
            vtemplate_df = vtemplate_df.loc[~vtemplate_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
            result["vteamplte"] = vtemplate_df

            yield result
    
    def data_generator(self, representation_l, val_representation_l=None):
        valFlag = False
        for i, representation_path in enumerate(representation_l):
            if val_representation_l is not None:
                vrepresentation_path = val_representation_l[i]
                vrep = np.load(vrepresentation_path)
                valFlag = True
            rep = np.load(representation_path)

            if valFlag:
                yield rep, vrep
            else:
                yield (rep, None)
    
    def get_label_by_cv(self,cvr, kv_df):
        template_df = cvr["template"]
        vtemplate_df = cvr["vtemplate"]

        if "query" in template_df.columns:
            template_df = template_df.join(kv_df.set_index("template"), on="query")
            template_df = template_df.join(kv_df.set_index("template"), on="template", lsuffix="_query",rsuffix="_template")
            label_ss = template_df.loc[:, "sf_query"] == template_df.loc[:, "sf_template"]

            vtemplate_df = vtemplate_df.join(kv_df.set_index("template"), on="query")
            vtemplate_df = vtemplate_df.join(kv_df.set_index("template"), on="template", lsuffix="_query",rsuffix="_template")
            vlabel_ss = vtemplate_df.loc[:, "sf_query"] == vtemplate_df.loc[:, "sf_template"]

            y = label_ss.to_numpy(dtype=np.int64)
            ey = vlabel_ss.to_numpy(dtype=np.int64)
        else:
            template_df = template_df.join(kv_df, on="template")
            vtemplate_df = vtemplate_df.join(kv_df, on="template")

            m_df = pd.concat([template_df.loc[:,"sf"], vtemplate_df.loc[:,"sf"]]).rename("sf")
            m_df = m_df.reset_index()
            num_class = len(m_df.groupby("sf"))
            m_df = m_df.assign(new_label=list(range(num_class)))
            template_df = template_df.join(m_df.set_index("sf"),on="sf")
            vtemplate_df = vtemplate_df.join(m_df.set_index("sf"),on="sf")

            y = template_df.loc[:, "new_label"].to_numpy(np.int64)
            ey = template_df.loc[:, "new_label"].to_numpy(np.int64)
        return y, ey

    
    def get_data_by_cv(self, cvr, rep, vrep,spec_df, vspec_df):
        template_df = cvr["template"]
        vtemplate_df = cvr["vtemplate"]

        if "query" in template_df.columns:
            # 需要调用sim
            # 因为这表示这是pairwise类型的检索结果
            template_df = template_df.join(spec_df,on="query")
            template_df = template_df.join(spec_df,on="template", lsuffix="_query",rsuffix="_template")
            template_df = \
                template_df.loc[:["row num_query","row num_template"]].rename(columns={"row num_query":"query","row num_template": "template"})
            X = self.similarity(pairwise_df = template_df, rep=rep)

            vtemplate_df = vtemplate_df.join(spec_df,on="query")
            vtemplate_df = vtemplate_df.join(spec_df,on="template", lsuffix="_query",rsuffix="_template")
            vtemplate_df = \
                vtemplate_df.loc[:["row num_query","row num_template"]].rename(columns={"row num_query":"query","row num_template": "template"})
            eX = self.similarity(vpairwise_df = vtemplate_df, vrep=vrep)            
        else:
            template_df = template_df.join(spec_df,on="template")
            vtemplate_df = vtemplate_df.join(vspec_df,on="template")

            idx = template_df.loc[:,"row num"].to_numpy(np.int64)
            vidx = template_df.loc[:, "row num"].to_numpy(np.int64)

            X = rep[idx]
            eX = vrep[vidx]
        
        return X, eX
        

    def cv_model_training(self, params, init_model, cv_df, spec_df, vspec_df, kv_df, 
        representaion_l, val_representaion_l, save_l, **kArgs):

        if kArgs.get("template_df") is not None:
            template_df = kArgs["template_df"]

        else:
            template_df = spec_df.reset_index().rename(columns={"index":"template"})
            
        
        if kArgs.get("vtemplate_df") is not None:
            vtemplate_df = kArgs["vtemplate_df"]
        else:
            vtemplate_df = vspec_df.reset_index().rename(columns={"index":"template"})

        for cvr in self.training_generator(cv_df=cv_df, template_df=template_df,vtemplate_df=vtemplate_df):
            y, ey = self.get_label_by_cv(cvr, kv_df)
            for rep, vrep in self.data_generator(representaion_l, val_representaion_l):
                X, eX = self.get_data_by_cv(cvr, rep, vrep, spec_df, vspec_df)


    def cli_cv_training(self,params,cv_df, spec_df, vspec_df, kv_df, conf_path,
        representaion_l, val_representaion_l,save_l, time_suffix=None,kernel=True):
        flag = False
        tunning = True
        valid_df = vspec_df.join(kv_df.set_index("template"))
        vlabel = valid_df.loc[:,"label"].to_numpy(np.int64)
        # vgroup = [valid_df.shape[0]]
        template_ss = pd.Series(spec_df.index,name="template")

        def gen_training(cv_df):
            count = 5
            for fold_num in cv_df.columns:
                if flag and count == 0:
                    break
                train_df = spec_df.loc[template_ss.loc[~template_ss.isin(cv_df.loc[:,fold_num])]]

                yield fold_num, train_df
                if flag:
                    count -= 1

        sel_ary = np.array([[False]] * 4)
        idx = [0,1]
        sel_ary[idx] = True
        for fold_num, train_df in gen_training(cv_df):
            train_df = train_df.join(kv_df.set_index("template"))
            # 重新给superfamily编号
            # 但是这里一开始忘记给vlabel重新编号了，因此，初步评估的结果并不准确需要重新评估
            k_df = train_df.groupby("sf").apply(lambda x: x.iloc[0,0]).reset_index()
            # print(k_df)
            k_df = k_df.reset_index().loc[:,["sf","index"]]
            # print(k_df)
            train_df = train_df.join(k_df.set_index("sf"),on="sf")
            train_df = train_df.rename(columns={"index":"new label"})
            classes = [x for x in train_df.groupby("label").groups.keys()] # classes还是用原来的标签，方便评估模型性能
            label = train_df.loc[:,"new label"].to_numpy(np.int64)
            
            # num_class = train_df.loc[:, "new label"].max()+1
            # params["num_class"] = num_class # SVM训练不需要知道num_class
            # label = label_binarize(label, classes=list(range(num_class))) # 用于方便计算roc_auc
            # vlabel = label_binarize(vlabel, classes=list(range(num_class))) # 用于方便计算roc_auc
            # group = [train_df.shape[0]]
            for i, representaion_path in enumerate(representaion_l):
                if not sel_ary[i]:
                    continue
                # print(vpairwise_df)
                val_representaion_path = val_representaion_l[i]
                rep = np.load(representaion_path)
                vrep = np.load(val_representaion_path)
                # print(vpairwise_df)
                idx = train_df.loc[:,"row num"].to_numpy(int)
                rep = rep[idx]

                # lgb_train, lgb_test = run_ltr.cv_training(sim,label,group,
                #     vsim,vlabel,egroup,params) # 用于测试
                
                printStart = "printf '数据集处理完毕\n'"
                sp.call(printStart,shell=True)
                saveDir = save_l[i]
                if tunning or time_suffix is not None:
                    #用于parameters tunning
                    savePath = os.path.join(
                        saveDir,"cv_0{}_{}".format(fold_num,time_suffix)
                    )
                else:
                    # 输出到optimal路径下，就表示参数都确定了
                    savePath = os.path.join(
                        saveDir,"cv_0{}".format(fold_num)
                    )
                # training_dataset_path = os.path.join(os.path.dirname(representaion_path),"multiclass.train")
                # validation_dataset_path = os.path.join(os.path.dirname(representaion_path),"multiclass.valid")

                # params["data"] = training_dataset_path
                # params["valid_data"] = validation_dataset_path
                # params["output_model"] = savePath
                
                # data = lgb.Dataset(rep,label,free_raw_data=False)
                # data.construct()
                # valid_data = lgb.Dataset(vrep,vlabel,free_raw_data=False)
                # valid_data.construct()

                # if os.path.exists(training_dataset_path):
                #     rm = "rm {}".format(training_dataset_path)
                #     sp.call(rm, shell=True)

                # data.save_binary(training_dataset_path)

                # if os.path.exists(validation_dataset_path):
                #     rm = "rm {}".format(validation_dataset_path)
                #     sp.call(rm, shell=True)

                # valid_data.save_binary(validation_dataset_path)

                # 归一化
                norms = self.row_norms(rep,[1,2,np.inf])
                norm = np.mean(norms, axis=1,keepdims=True) # 这里之前没用1/norms而是直接用了norms，但并不是没有效果
                rep /= norm

                norms = self.row_norms(vrep,[1,2,np.inf])
                norm = np.mean(norms, axis=1,keepdims=True) # 这里之前没用1/norms而是直接用了norms，但并不是没有效果
                vrep /= norm
                printStart = "printf '归一化完毕\n'"
                sp.call(printStart,shell=True)

                # # 保存为对应的格式用于LightGBM的训练
                # rep = np.hstack([label.reshape((-1,1)),rep])
                # vrep = np.hstack([vlabel.reshape((-1,1)),vrep])

                # print(rep.shape)
                # print(vrep.shape)
                # forceFlag = True
                # if forceFlag or not os.path.exists(training_dataset_path):
                #     self.construct_dataset(rep,training_dataset_path)
                # if forceFlag or not os.path.exists(validation_dataset_path):
                #     self.construct_dataset(vrep,validation_dataset_path)
                
                # # 再把params保存到conf_path中
                # self.save_conf(params, conf_path)

                # printStart = "printf '开始训练\n'"
                # sp.call(printStart,shell=True)

                # # 命令行调用
                # training = "lightgbm config={}".format(conf_path)
                # sp.call(training,shell=True)
                
                # # 调用sklearn的GradientBoostingClassifier
                # clf = GradientBoostingClassifier(**params)
                # printStart = "printf '开始训练\n'"
                # sp.call(printStart,shell=True)
                # with joblib.parallel_backend('dask'):
                #     clf.fit(rep, label)
                
                # printStart = "printf '利用验证集评估\n'"
                # sp.call(printStart,shell=True)
                # score = clf.decision_function(vrep)
                
                # auc_mu = self.auc_mu(score, vlabel, classes)
                # print("auc: {}".format(auc_mu))

                # with open(savePath, "wb") as h:
                #     cPickle.dump(clf, h)

                if kernel:
                    printStart = "printf '使用kernel svm\n'"
                    sp.call(printStart,shell=True)
                    self.kernel_svm(params, rep, label, vrep,vlabel, classes,savePath)
                else:
                # 调用linearsvm
                    printStart = "printf '使用linear svm\n'"
                    sp.call(printStart,shell=True)
                    self.sklearn_svm(params, rep, label, vrep,vlabel, classes,savePath)
    
    def sklearn_svm(self,params, X,y, eX, ey, classes, savePath):
        clf = LinearSVC(**params)
        printStart = "printf '开始训练\n'"
        sp.call(printStart,shell=True)

        clf.fit(X, y)

        printStart = "printf '保存模型到{}\n'".format(savePath)
        sp.call(printStart,shell=True)
        with open(savePath, "wb") as h:
            cPickle.dump(clf, h)

        printStart = "printf '开始评估\n'"
        sp.call(printStart,shell=True)
        score = clf.decision_function(eX)

        auc_mu = self.auc_mu(score, ey, classes)
        printStart = "printf {}".format(auc_mu)
        sp.call(printStart,shell=True)

    def kernel_svm(self,params, X,y, eX, ey, classes, savePath):
        clf = SVC(**params)
        printStart = "printf '开始训练\n'"
        sp.call(printStart,shell=True)
        
        clf.fit(X, y)

        printStart = "printf '保存模型到{}\n'".format(savePath)
        sp.call(printStart,shell=True)
        with open(savePath, "wb") as h:
            cPickle.dump(clf, h)

        printStart = "printf '开始评估\n'"
        sp.call(printStart,shell=True)
        score = clf.decision_function(eX)

        auc_mu = self.auc_mu(score, ey, classes)
        printStart = "printf {}".format(auc_mu)
        sp.call(printStart,shell=True)            

    def save_conf(self, params, conf_path):
        with open(conf_path, "w") as h:
            for key in params:
                line = "{} = {}\n".format(key,params[key])
                h.write(line)
    
    def construct_dataset(self, data, dataset_path):
        r, c = data.shape
        format = "\t".join(["%.4f"]*(c-1))
        with open(dataset_path,"w") as h:
            for i in range(r):
                x = data[i]
                label = int(x[0])
                feature = tuple(x[1:])
                f_str = format % feature
                line = "{}\t{}\n".format(label,f_str)
                h.write(line)
            print(label)
    
    def row_norms(self, mat, ords,axis=1):
        norms = np.hstack(
            [
                np.linalg.norm(mat,ord,axis=axis,keepdims=True)\
                    for ord in ords
            ]
        )
        return norms
    
    def auc_mu(self, y_pred, y_true, classes):
        y_true = label_binarize(y_true, classes=classes)
        # fpr = dict()
        # tpr = dict()
        # roc_auc = []
        # for i in range(len(classes)):
        #     fpr[i], tpr[i], _ = roc_curve(y_true[:, i], y_pred[:, i])
        #     if not np.isnan(tpr[i]).any():
        #         try:
        #             roc_auc.append(auc(fpr[i], tpr[i]))
        #         except:
        #             if 1 in tpr[i]:
        #                 roc_auc.append(1.)
        #             else:
        #                 roc_auc.append(0.)
        # auc_score = np.array(roc_auc).mean()
        y_pred = y_pred.reshape((-1,))
        y_true = y_true.reshape((-1,))

        auc_score = roc_auc_score(y_true,y_pred)
        return auc_score


    # def LTR_cluster(self, clf, s2c, rep, vrep)