import numpy as np
import pandas as pd
import tensorflow as tf
import os
from six.moves import cPickle


class TriadicClosurePrinciple(object):
    def __init__(self, topn=-1):
        self.topn = topn
    
    # r是一个n_class大小的tensor, ltr是一个n_template*n_class大小的tensor
    # 对每一种class，取属于该class的序列所对应的ltr排序结果
    # 这里的排序结果, 包括r和ltr都是seq2cluster类型的排序结果，
    # 即排序依据是基于序列与superfamily之间的关系来排序的
    # m_df是包含template, row num, new_row_num, sf, label, new_label这几列的dataframe
    def rerank(self, r, ltr, m_df, topn=None):
        if topn is None:
            topn = self.topn

        r = self.normalization(r,0)
        def cal_dist(p_df):
            indices = tf.constant(
                p_df.loc[:,"new_row_num"].to_numpy(),
                tf.int32
            )
            s2c = tf.gather_nd(ltr,tf.reshape(indices,(-1,1)))
            s2c = self.normalization(s2c,1)
            # 这里用JS或者KL会更好
            dcon = self.JS_divergence(r+1,s2c+1) # 因为通过归一化后，r以及s2c的范围在(-1,1)
            d = tf.math.reduce_mean(
                tf.math.reduce_sum(dcon,1)
                ,0
            ) #存在多个向量先求和，再取平均值
            return d

        # sf_idxs = tf.argsort(r, axis=0, direction="DESCENDING")
        # sf_df = m_df.set_index("new_label").loc[sf_idxs]
        # up_df = sf_df.groupby(level=0).apply(cal_dist)
        # sf_idxs = list(range(r.shape[0]))
        up_df = m_df.groupby("new_label").apply(cal_dist)
        nr = tf.constant(up_df.to_numpy(),tf.float32)

        # nr = tf.tensor_scatter_nd_update(r,sf_idxs,updates)

        return nr
    
    def rerankt(self, r, ltr, m_df):
        r = self.normalization(r,0)
        ltr = self.normalization(ltr, 1)

        # 先求出所有query与template之间的JSD
        # dcon大小为training_sample * n_class大小
        dcon = self.JS_divergence(r+1, ltr+1)
        # dcon大小为training_sample
        dcon = tf.math.reduce_sum(dcon,axis=1)
        # print(dcon)

        def cal_dist(p_df):
            indices = tf.constant(
                p_df.loc[:,"new_row_num"].to_numpy(),
                tf.int32
            )
            ds2c = tf.gather_nd(dcon,tf.reshape(indices,(-1,1)))
            d = tf.math.reduce_mean(ds2c,0) #存在多个向量先求和，再取平均值
            return d

        # 然后根据superfamily分类进行处理
        up_df = m_df.groupby("label").apply(cal_dist)
        nr = tf.constant(up_df.to_numpy(),tf.float32)
        nr = tf.reshape(nr, r.shape)
        return nr
    
    # rt是test_sample * n_class大小
    # ltr是training_sample * n_class大小
    # m_df包含template, row num, new_row_num, sf, label这几列
    def rerank_t2t(self, rt, ltr, m_df):
        rt1 = tf.transpose(tf.reshape(rt,rt.shape+[1]),[0,2,1])
        rt1 = self.normalization(rt1,2)
        ltr = self.normalization(ltr,1)
        jsd_con = self.JS_divergence(rt1+1,ltr+1)

        # jsd_con是test_sample * training_sample * n_class大小
        # dcon是test_sample * training_sample大小
        dcon = tf.math.reduce_sum(jsd_con,axis=2)
        mg = m_df.groupby("label")
        values = mg.groups.values()

        def cluster_sim(idx):
            indices = tf.constant(
                m_df.loc[idx,"new_row_num"].to_numpy(),
                tf.int32
            )

            # cc是test_sample * k大小，k表示部分training序列的个数
            cc = tf.transpose(
                tf.gather_nd(tf.transpose(dcon),
                tf.reshape(indices,(-1,1))
                )
            )

            # csim是test_sample*1大小
            csim = tf.reduce_mean(cc,axis=1,keepdims=True)
            return csim
        
        # sim_list由n_class个大小为test_sample*1的tensor
        sim_list = list(map(cluster_sim,values))
        # nrt是大小为test_sample*n_class大小的tensor
        nrt = tf.transpose(tf.concat(sim_list,axis=1))

        return nrt


    def KL_divergence(self, t1, t2):
        kld = t1 * tf.math.log(t1 / t2)
        return tf.abs(kld)

    def JS_divergence(self, t1, t2):
        m = 1 / 2 * (t1+t2)
        kld1 = self.KL_divergence(t1,m)
        kld2 = self.KL_divergence(t2,m)
        jsd = 1 / 2 * (kld1 + kld2)
        return jsd

    def dist(self, t1, t2):
        jsd = self.JS_divergence(t1,t2)
        d = tf.math.sqrt(2*jsd)
        return d

    def normalization(self, con, axis):
        cnorms = self.norms(con,[1,2,np.inf],axis)
        cnorm = tf.math.reduce_mean(cnorms,axis,keepdims=True)
        con /= cnorm

        return con

    def norms(self, tensor, ords, axis):
        norm_list = [
            tf.norm(tensor,ord=x,axis=axis,keepdims=True)\
                for x in ords
        ]
        norms = tf.concat(norm_list,axis=axis)
        return norms