import lightgbm as lgb
import numpy as np
import pandas as pd
import tensorflow as tf
import subprocess as sp
import re
import os
from sklearn.preprocessing import label_binarize

class LambdaMART(object):
    def load_training_data(self, representation_path):
        if os.path.exists(representation_path):
            rep = tf.constant(np.load(representation_path),tf.float32)

            # 表示这是n_sample * n_class的tensor
            # 要转为n_sample * n_class * 1的tensor
            if len(rep.shape) == 2:
                rep = tf.reshape(rep, rep.shape +[1])
            
            assert len(rep.shape) == 3
            return rep
        else:
            return None

    def get_training_set(self, representation_l):
        constant_list = [self.load_training_data(x) for x in representation_l]

        # 保证并非全为None
        assert constant_list != [None]*len(constant_list)

        # 获取首个不为None的constant的size    
        con_shape = None  
        for con in constant_list:
            if con is not None:
                con_shape = con.shape
                break
        
        if con_shape is not None:
            for i in range(len(constant_list)):
                if constant_list[i] is None:
                    constant_list[i] = tf.zeros(con_shape)

        # 这是以前的方法
        # con = tf.stack(constant_list,axis=0) # 得到k*n_sample*n_class大小的tensor
        # con = tf.transpose(con,[1,2,0]) # 得到n_sample*n_class*k大小的tensor
        con = tf.concat(constant_list,axis=2) # 得到n_sample*n_class*k大小的tensor

        return con

    def normalization(self, con, axis):
        cnorms = self.norms(con,[1,2,np.inf],axis)
        cnorm = tf.math.reduce_mean(cnorms,axis,keepdims=True)
        con /= cnorm

        return con
    
    # 由于是按字典序排序的，如果有缺失很容易补全
    # pairwise_df是检索结果
    # 通过query_ser补全检索结果，原本没有的检索结果，用0代替
    # query_df已经按字典序排序了
    def completement_test_set(self, query_ser, pairwise_df, con):
        cshape = con.shape
        ncshape = [query_ser.shape[0]] + cshape[1:]
        zcon = tf.zeros(ncshape)

        # 编号
        query_df = pd.DataFrame({"template":query_ser,"row num":list(range(query_ser.shape[0]))})

        idx = query_df.loc[query_df.loc[:,"template"].isin(pairwise_df.loc[:,"query"]),"row num"]

        indices = tf.constant(idx.to_numpy(),tf.int32)
        indices = tf.reshape(indices,(-1,1))

        zcon = tf.tensor_scatter_nd_update(zcon,indices,con)
        return zcon
    
    # template_df是符合交叉验证划分结果的
    def get_label(self, kv_df, kv_i_df, template_df,keepdims=False):
        train_df = template_df.join(kv_df.set_index("template"),on="template")
        # 根据字典序排序
        # train_dfrow num的顺序排列
        # train_df = train_df.sort_values(by="template")# 不应该基于template排序
        train_df = train_df.sort_values(by="row num")
        classes = tf.constant(train_df.loc[:,"label"].to_numpy(),tf.int32) # classes还是用原来的标签，方便评估模型性能

        # template_df有可能是训练集的template也有可能是验证集的template
        # 所以用于比较的label，要通过kv_i_df得到

        # 注释的部分, 是以前的程序
        # kv_i_df = kv_df.loc[~kv_df.loc[:,"template"].isin(cv_df.loc[:,fn_col])]
        # kv_i_df = kv_i_df.sort_values(by="label")

        label = tf.constant(list(kv_i_df.groupby("label").groups.keys()),tf.int32)

        # label = tf.constant(sf_df.loc[:,"label"].to_numpy(),tf.int32)

        classes = tf.reshape(classes,(-1,1))
        label = tf.reshape(label,(1,-1))
        y_true = tf.cast(classes == label,tf.int32)
        if not keepdims:
            y_true = tf.reshape(y_true,(-1,))
        return y_true
    
    # con为n_samples*n_class*k
    # indices的目的是把con转为n_samples*n_class*n_feature, n_feature < k
    # 比如k=6，indices=[[0,1],[2,3,4,5]], 那么n_feature=2
    def merge_methods(self, con,indices):
        # # 列归一化
        # con = self.normalization(con,1)

        c = tf.transpose(con,[2,0,1])

        # 假设idx为k_1
        def average(idx):
            # 得到k_1*n_samples*n_class
            m = tf.gather_nd(c, tf.reshape(idx,(-1,1)))
            # 得到1*n_samples*n_class
            m1 = tf.math.reduce_mean(m,axis=0,keepdims=True)
            return m1

        # new_list中包含了n_feature个大小为1*n_samples*n_class的tensor
        new_list = list(
            map(average,indices)
        )

        # 得到n_feature*n_samples*n_class
        new_con = tf.concat(new_list,axis=0)
        # 得到n_samples*n_class*n_feature
        new_con = tf.transpose(new_con,[1,2,0])

        return new_con
        
    
    def load_conf(self, parameters_path):
        spliter = re.compile(r"\s*=\s*")
        comment_parser = re.compile(r"^#")
        p_list = []
        with open(parameters_path,"r") as h:
            for l in h:
                l = l.rstrip()
                r = spliter.search(l)
                cr = comment_parser.search(l)
                if cr is not None:
                    continue
                if r is not None:
                    p_list.append(spliter.split(l))
        params = dict(p_list)
        return params
    
    def save_conf(self, params, parameters_path):
        with open(parameters_path,"w") as h:
            for key, value in params.items():
                line = "{} = {}\n".format(key,value)
                h.write(line)

    def cli_train(self, params_path):
        lightgbm = "lightgbm config={}".format(params_path)
        sp.call(lightgbm,shell=True)
    
    def train(self, params, **kArgs):
        kArgs["params"] = params
        booster = lgb.train(**kArgs)
        return booster
    
    def norms(self, tensor, ords, axis):
        norm_list = [
            tf.norm(tensor,ord=x,axis=axis,keepdims=True)\
                for x in ords
        ]

        norms = tf.concat(norm_list,axis=axis)
        return norms