import numpy as np
import pandas as pd
import tensorflow as tf
from itertools import combinations
from six.moves import cPickle
import subprocess as sp
import tensorflow as tf
import time


# 提供一个pairwise格式的排序列表，文件为csv格式
# 包含query, template, prob, evalue这几列（其中prob可以换成score)
class ReRanker(object):
    # pairwise_df表示符合交叉验证划分的检索结果
    # 注意kv_df也符合按照交叉验证划分的结果
    def expand_query(self, pairwise_df, kv_df):
        pairwise_df = pairwise_df.join(kv_df.set_index("template"),on="query")
        return pairwise_df
    
    # p_df表示符合交叉验证划分的某一个query下的检索结果
    # 注意kv_df也符合按照交叉验证划分的结果
    def expand_template(self, p_df, kv_df):
        m_df = kv_df.join(p_df.set_index("template"),on="template",lsuffix="_template",rsuffix="_query")
        return m_df
    
    def construct_graph_from(self, prob_con=None, size=None):
        assert prob_con is not None or size is not None

        if prob_con is not None:
            epsilon = 1 / tf.norm(prob_con,1, 0) # 用于计算KL时调整，防止出现nan值
            # DFG
            rp_con = tf.reshape(1 / (prob_con+epsilon),(1,-1))
            prob_con = tf.reshape(prob_con,(-1,1))
            gcon = tf.matmul(prob_con+epsilon,rp_con)
            gcon = tf.math.log(gcon)
            gcon *= (prob_con+epsilon)
            gcon = tf.abs(gcon)
            gcon = gcon / (1+gcon)
            boolean_mask = prob_con > tf.reshape(prob_con,(1,-1))
            indices = tf.where(boolean_mask)
            updates = tf.zeros(indices.shape[0],tf.float32)
            gcon = tf.tensor_scatter_nd_update(gcon,indices,updates)
        else:
            gcon = tf.zeros(size,tf.float32)
        return gcon
    
    def add_query(self, gcon, prob_con, strategy="dfg"):
        if strategy == "dfg":
            p_max = tf.math.reduce_max(prob_con)
            updates = tf.reshape(1 / (1 + p_max*tf.math.log(p_max/prob_con)),(1,-1))
            gcon = tf.concat([gcon,updates],axis=0)
        elif strategy == "normal":
            N = prob_con.shape[0]
            gcon = tf.concat([gcon,tf.zeros((1,N))],axis=0)
            indices = tf.argsort(prob_con,axis=0,direction='DESCENDING')
            size = indices.shape[0]
            qs = N * tf.ones((size,1),tf.int32) # query在最后一行
            # ids
            indices = tf.concat([qs,indices],axis=1)
            updates = 1 / np.array(list(range(1,size+1)),np.float32) #1 / (pos + 1) 这里pos从0开始
            gcon = tf.tensor_scatter_nd_update(gcon,indices, updates)
        return gcon


    # m_df表示检索结果, 这个检索结果是先后经过expand_query和expand_template的
    # spec_df表示序列对应的行下标（列下标）
    # kv_df表示序列对应的超家族信息
    # 注意m_df和kv_df都符合按照交叉验证划分的结果
    # 返回的类型是tensor，这样方便整个多个排序结果
    # 因为tensor类型也可以通过np.save的方式保存成npy的格式
    
    def pagerank(self,m_df, spec_df, alpha=0.85, iterations=2, threshold=0.8):
        # s = time.time()

        N = m_df.shape[0]

        m_df = m_df.join(spec_df, on="template")
        m_df = m_df.astype({"row num":int})
        m_df = m_df.sort_values(by="row num")
        # 重新编号
        m_df = m_df.assign(new_row_num=list(range(N)))

        prob_con = tf.constant(m_df.loc[:,"prob"].to_numpy(),tf.float32)
        indices = tf.where(tf.math.is_nan(prob_con))
        updates = tf.zeros(indices.shape)
        prob_con = tf.tensor_scatter_nd_update(tf.reshape(prob_con,(-1,1)), indices, updates)

        # 当排序结果较好的时候, 就不构建DFG
        # 排序结果的好坏可以，通过查询最大值
        # 因此初始排序结果最好能反映真实的远同源性
        # 比如hhblits的prob值， score值
        # evalue值要注意，evalue越小表示，结果越具有可信度
        gcon = self.construct_graph_from(size=(N,N)) # 构建全为0的graph

        # 同一个superfamily共享一个打分
        from itertools import combinations

        gcon1 = gcon
        indices = []
        for _, df in m_df.groupby("sf_template"):
            idx = df.loc[:, "new_row_num"].to_list()
            comb_idx = list(combinations(idx,2)) +\
                list(combinations(idx[::-1],2))
            comb_idx = sorted(comb_idx)
            indices += comb_idx
        indices = tf.reshape(tf.constant(indices,tf.int32),(-1,2))
        updates = tf.ones(indices.shape[0],tf.float32)
        gcon1 = tf.tensor_scatter_nd_update(gcon1,indices,updates)

        # 添加query到graph
        gcon1 = self.add_query(gcon1, prob_con, strategy="dfg") # 使用dfg策略结果会更好

        # 补全graph, 初始化排序结果
        alpha = 0.85
        N, _ = m_df.shape
        N += 1
        gcon1 = tf.concat([gcon1,tf.zeros((N,1))],axis=1)
        # 给graph添加上自环
        # print(gcon1.shape)
        indices = tf.reshape(tf.constant(list(range(N)),tf.int32),(-1,1))
        updates = tf.ones(indices.shape[0])
        indices = tf.concat([indices,indices],axis=1)
        # gcon1 += tf.eye(N)
        gcon1 = tf.tensor_scatter_nd_update(gcon1,indices,updates) #添加自环

        # # graph归一化
        # gnorms = self.norms(gcon1,[1,2,np.inf],1)
        # gnorm = tf.math.reduce_mean(1 / gnorms, 1)
        # gnorm = tf.reshape(gnorm, (-1,1))
        # gcon1 = gcon1 * gnorm

        # graph归一化
        # 归一化方法修改一下
        gnorms = self.norms(gcon1,[1,2,np.inf],1)
        gnorm = tf.math.reduce_mean(gnorms, 1,keepdims=True)
        gcon1 /= gnorm

        # # R = (2+N) / (3*N) * tf.ones((N,1)) 
        R = tf.zeros((N,1))
        # query要赋值1
        R = tf.tensor_scatter_nd_update(R,[[N-1]],[[1]])
        gcon1 = alpha * tf.transpose(gcon1) + (1-alpha) * 1 / N

        # # 使用pagerank
        # R = tf.reshape(tf.concat([prob_con,[[100.]]],axis=0), (N,1)) #这种初始化能得到更好的结果
        iterations = 2
        for _ in range(iterations):
            R = tf.matmul(gcon1, R)
            R = tf.reshape(R,(-1,1))
            vnorms = self.norms(R,[1,2,np.inf],0)
            vnorm = tf.reduce_mean(1/vnorms, 0)
            R = R * vnorm
            R = tf.reshape(R,(-1,1))

        # e = time.time()
        # print("pagerank time: {}".format(e-s))
        return R
    
    def norms(self, con, ords, axis):
        norm_list = [
            tf.norm(con,ord,axis,keepdims=True) \
                for ord in ords
        ]

        return tf.concat(norm_list,axis)
    
    def get_pairwise_df(self, srch_df, fold_num, cv_df):
        srch_df = srch_df.loc[srch_df.loc[:,"query"].isin(cv_df.loc[:,fold_num])]
        srch_df = srch_df.loc[~srch_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]

        return srch_df
    
    def get_key2value_df(self, key2value_df, fold_num, cv_df, spec_df):
        kv_df = spec_df.join(key2value_df.set_index("template")).reset_index().rename(columns={"index":"template"})
        kv_df = kv_df.loc[~kv_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
        return kv_df.loc[:,["template","sf","label"]]

