# 用于转换特征，通过LR，SVM或者构造DFG等方式处理特征，使得特征能够用于LTR的训练

# 构建DFG
# 注意排序列表为pairwise打分，所有pairwise（排序方法能够得到的结果）放在一个文件中，文件格式为csv

# pairwiseFilePath不能为空
# 构建DFG只用三种值,prob, evalue, score，因此，不同的文件中的column要一致，一定要包含这3种名称

# 内部设计较大的运算，返回的中间结果都是tensor类型，只有在最后保存的时候，会使用numpy来保存

import pandas as pd
import numpy as np
import tensorflow as tf
import lightgbm as lgb
from six.moves import cPickle
import os
import time
import re
import subprocess as sp
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.svm import LinearSVC
from sklearn.preprocessing import label_binarize
from sklearn.metrics import roc_curve, auc
from dask.distributed import Client
import joblib

class DFG(object):
    def __init__(self, queryName, *pairwiseFilePath):
        self.qN = queryName
        self.pF_path = pairwiseFilePath
        # self.df_list = self.fetchRanking(queryName,*pairwiseFilePath) # 获取对应的排序结果
        self.df_list = self.read_ranking(*pairwiseFilePath)
    
    def fetchRanking(self, queryName, *pairwiseFilePath):
        def gen(pF_path_list):
            for pairwiseFilePath in pF_path_list:
                pw_df = pd.read_csv(pairwiseFilePath)
                result_df = pw_df.loc[pw_df.loc[:,"query"] == queryName,["template","prob","score"]]

                # 进行归一化
                r_ary = result_df.loc[:,["prob","score"]].to_numpy(np.float32)
                r_norms = self.row_norms(r_ary,[1,2,np.inf])
                r_norm = np.average(1 / (1+r_norms),0).reshape((-1,))
                r_ary = r_ary * r_norm
                result_df.loc[:,["prob","score"]] = r_ary
                yield result_df.reset_index(drop=True)
        result_df_list = list(gen(pairwiseFilePath))
        return result_df_list
    
    def read_ranking(self, *pairwiseFilePath):
        def gen(pF_path_list):
            for pF_path in pF_path_list:
                pw_df = pd.read_csv(pF_path)
                yield pw_df
        return list(gen(pairwiseFilePath))
    
    def get_query_ranking(self, queryName, dataframeList=None):
        if dataframeList is None:
            dataframeList = self.df_list
        def gen(dataframeList):
            for pw_df in dataframeList:
                result_df = pw_df.loc[pw_df.loc[:,"query"] == queryName,["template","prob","score"]]
                # 进行归一化
                r_ary = result_df.loc[:,["prob","score"]].to_numpy(np.float32)
                r_norms = self.row_norms(r_ary,[1,2,np.inf])
                r_norm = np.average(1 / (1+r_norms),0).reshape((-1,))
                r_ary = r_ary * r_norm
                result_df.loc[:,["prob","score"]] = r_ary
                yield result_df.reset_index(drop=True)
        return list(gen(dataframeList))
    
    # dataframeList里是不同方法下的pairwise打分结果
    def construct(self, dataframeList):
        # 因为，所有dataframe的index都是一致的，因此，只要注意顺序，不用额外制作一个dataframe来表示这种关系
        def get(dataframe):
            mat = dataframe.loc[:,["prob","score"]].to_numpy(np.float32)
            # con = tf.constant(mat,dtype=tf.float64) # 策略1
            con = tf.constant(mat)
            
            # 策略1，使用生成器
            # r,_ = con.shape
                        
            # def cal_KL(r):
            #     for i in range(r):
            #         tmp = np.log(con[i] / con)
            #         r = tf.matmul(tmp,tf.reshape(con[i],(-1,1)))
            #         yield r
            
            # KL_con = tf.transpose(tf.stack(list(cal_KL(r))))

            # 策略2，使用tf.map_fn
            def cal_KL(r_x):
                # 这里有可能出现inf值
                tmp = tf.math.log(r_x / con)
                r = tf.abs(tf.matmul(tmp, tf.reshape(r_x,(-1,1))))
                return tf.reshape(r,(-1,))

            # 注意这里可能会出现inf值
            KL_con = tf.map_fn(cal_KL,con)

            # 通过x / (1+x)进行概率化
            return KL_con / (1+KL_con)

        kl_list = [get(df) for df in dataframeList]
        try:
            KL_con = tf.math.reduce_mean(kl_list,axis=0)
        except Exception as e:
            print(e)
            print(kl_list)
            print(self.qN)
            for df in kl_list:
                print(df.shape)

            # print(kl_list)
        # 由于出现na值
        # 以上这些问题都是一个问题，因此，尝试用tf.float64来解决
        # 策略2是，把nan值替换为1，因为inf/(1+inf)为1，同时，由于分母不会为-1，因此，nan值出现只可能有这种情况
        
        def update(r_x):
            idx = tf.where(tf.math.is_nan(r_x))
            size = idx.get_shape()[0]
            up_value = tf.ones((size,))
            return tf.tensor_scatter_nd_update(r_x, idx, up_value)
        
        KL_con = tf.map_fn(update,KL_con)

        r,c = KL_con.get_shape()
        # 添加自环
        KL_con += tf.eye(r,c)
        # ones_con = tf.reshape(tf.ones(c,dtype=tf.float64),(-1,1)) # 如果是策略1，类型要保持一致
        ones_con = tf.reshape(tf.ones(c),(-1,1))
        ranking_con = tf.reshape(tf.matmul(KL_con,ones_con),(-1,))


        # 由于template都是一样的（经过consistent处理）
        # 任意取一个即可
        df = dataframeList[0]
        # print(df.shape,ranking_con.get_shape())
        ranking_df = df.loc[:,["template"]].assign(prob=ranking_con.numpy())
        ranking_df = ranking_df.assign(query=self.qN)
        ranking_df = ranking_df.loc[:,["query","template","prob"]]
        return ranking_df

    
    # 保持不同dataframe长度一致
    # 因为不同的方法不一定检索出来的结果一致，因此需要，处理一下
    def consistent(self, spec_df, dataframeList):
        std_df = pd.concat(dataframeList).reset_index(drop=True)
        stdDf_g = std_df.groupby("template")
        idx = stdDf_g["prob"].idxmax().reset_index().loc[:,"prob"]
        std_df = std_df.loc[idx].reset_index(drop=True)
        # 注意dataframe中可能存在重复的template
        # 这个应当提前去除
        for i in range(len(dataframeList)):
            df = dataframeList[i]
            # df_g = df.groupby("template")
            # idx = [x[0] for x in df_g.groups.values()]
            # df = df.loc[idx].reset_index(drop=True)
            idx = ~std_df.loc[:,"template"].isin(df.loc[:,"template"])
            diff_df = std_df.loc[idx]
            m_df = pd.concat([df, diff_df])
            
            # 保证顺序和特征矩阵中的行下标一致
            m_df = m_df.join(spec_df,on="template").sort_values(by="row num").reset_index(drop=True) # 这样使得template的位置都是一致的
            dataframeList[i] = m_df
        return dataframeList

        
    def row_norms(self, mat,ord_list):
        return np.vstack(
            [
                np.linalg.norm(mat,ord,0,keepdims=True) \
                    for ord in ord_list
            ]
        )
    
    def get_dataframe_list(self):
        return self.df_list

    # 用于改变参数
    def set_params(self,**kArgs):
        if kArgs.get("queryName") is not None:
            self.qN = kArgs["queryName"]
        if kArgs.get("pairwiseFilePath") is not None:
            self.pF_path = kArgs["pairwiseFilePath"]
            self.df_list = self.read_ranking(*self.pF_path) # 获取对应的排序结果

# 通过线性分类器或者线性回归来得到query与template的打分或者query与superfamily的打分
# 
class LabelModel(object):    
    # 这里返回一个默认的分类模型，之后可以通过set_params api进行参数的设置
    def classifier(self):
        return lgb.LGBMClassifier()
    
    # 通过pairwise_df指定的pairwise计算similarity
    # pairwise_df包含query, template
    # 每一列的数字代表rep矩阵的行下标
    # 如果vrep为None就是，那么pairwise_df对应的行下标就都在rep中
    # 反之，则query在vrep中
    def pairwise_sim(self, pairwise_df, rep,vrep=None):
        if type(rep) == np.ndarray:
            ron = tf.constant(rep,tf.float32)
        else:
            ron = rep
        
        if vrep is not None:
            if type(vrep) == np.ndarray:
                vron = tf.constant(vrep,tf.float32)
            else:
                vron = vrep   
        
        lidx = pairwise_df.loc[:,"query"].to_numpy().reshape((-1,1))
        ridx = pairwise_df.loc[:,"template"].to_numpy().reshape((-1,1))
        
        if vrep is None:
            dis = tf.gather_nd(ron,lidx) - tf.gather_nd(ron, ridx)
        else:
            # print(vron)
            # print(lidx)
            # print(ron)
            # print(ridx)
            dis = tf.gather_nd(vron,lidx) - tf.gather_nd(ron, ridx)

        dis = tf.abs(dis)

        maximum = tf.linalg.norm(dis, ord=np.inf,axis=1,keepdims=True)
        log = tf.math.log(
            tf.linalg.norm(
                dis,ord=2,axis=1,keepdims=True
            )
        )

        log /= tf.math.log(3.)

        sim = tf.concat([dis, maximum, log],axis=1)
        return sim    

    # 确实可以直接使用特征，进行分类，但LightGBM是类似MART的模型，因此，是可以建立相似性特征来训练
    # 这里参考Fold-LTR-TCP的处理方法
    # |x_1 - x_2|, max(|x_1 - x_2|), log_a(sum((x_1-x_2)^2))
    # 假设特征原本维度为n, 那么这里可以得到n+2维相似性特征
    # representation表示原本的特征，n_sameples * n_features
    # 会输出(n_sampeles^2) * (n_features + 2)的特征矩阵
    # 每行代表的就是pairwise的结果，因为就是组合，所以，直接把template name按照combination，k=2输出
    # 就能表示每行的具体含义，同时，也方便给出对应的label
    # 这就是为什么这里返回Tensor的原因，就是为了方便操作
    def similarity(self, representaion):
        if type(representaion) == np.ndarray:
            rep_con = tf.constant(representaion,tf.float32)
        else:
            rep_con = representaion

        tmp = tf.map_fn(lambda x: tf.abs(x-rep_con),rep_con)
        a, b, c = tmp.get_shape()
        dis_con = tf.reshape(tmp,(a*b,c))
        # mask = tf.map_fn(lambda x:all(x),tf.greater(dis_con,0)) # 去掉全部为0的行，那表示的是行向量减去自身的结果
        # dis_con = tf.boolean_mask(dis_con,mask)
        max_con = tf.linalg.norm(dis_con,ord=np.inf,axis=1,keepdims=True)
        mask = tf.reshape(tf.greater(max_con,0),(-1,)) # 因为之前dis_con算的是abs，因此，max中未大于0的就是行向量减去自身的结果
        dis_con = tf.boolean_mask(dis_con,mask)
        max_con = tf.boolean_mask(max_con,mask)
        log_con = tf.math.log(
            tf.linalg.norm(
                dis_con,ord=2,axis=1,keepdims=True
            )
        )
        log_con /= tf.math.log(3.0) # 计算以3为底的对数

        sim_con = tf.concat([dis_con,max_con,log_con],axis=1)
        return sim_con
    
    # 计算lrep每行与rrep的sim
    def sim_between(self, lrep,rrep):
        if type(lrep) == np.ndarray:
            lr_con = tf.constant(lrep,tf.float32)
        else:
            lr_con = lrep

        if type(rrep) == np.ndarray:
            rr_con = tf.constant(rrep,tf.float32)
        else:
            rr_con = rrep

        tmp = tf.map_fn(lambda x: tf.abs(x-rr_con),lr_con)
        a,b,c = tmp.get_shape()
        dis_con = tf.reshape(tmp,(a*b,c))
        # mask = tf.map_fn(lambda x:all(x),tf.greater(dis_con,0)) # 去掉全部为0的行，那表示的是行向量减去自身的结果
        # dis_con = tf.boolean_mask(dis_con,mask)
        max_con = tf.linalg.norm(dis_con,ord=np.inf,axis=1,keepdims=True)
        mask = tf.reshape(tf.greater(max_con,0),(-1,)) # 因为之前dis_con算的是abs，因此，max中未大于0的就是行向量减去自身的结果
        dis_con = tf.boolean_mask(dis_con,mask)
        max_con = tf.boolean_mask(max_con,mask)
        max_con = tf.linalg.norm(dis_con,ord=np.inf,axis=1,keepdims=True)
        log_con = tf.math.log(
            tf.linalg.norm(
                dis_con,ord=2,axis=1,keepdims=True
            )
        )
        log_con /= tf.math.log(3.0) # 计算以3为底的对数

        sim_con = tf.concat([dis_con,max_con,log_con],axis=1)
        return sim_con        


    # spec_df表示template name对应representaion matrix的位置
    # 即第几行
    # 这是pandas.DataFrame类型，其中必然包含了template一列
    # 但是注意原有的文件中，template name是作为index的，因此需要提前修改
    # 这个函数则输出组合结果，1表示属于同一superfamily，0表示不属于，结果用tensor表示
    # key2valueDictFilePath表示fas对应的信息，比如这里就是sequence 对应superfamily type
    def pairwiseTemplate(self, spec_df,key2value_dict):
        # df = pd.read_csv(csvFilePath,index_col=0)
        df = spec_df

        # with open(key2valueDictFilePath,"rb") as h:
        #     key2value_dict = cPickle.load(h)

        # 通过笛卡尔积建立组合
        # 先建立一个全为1的key
        df = df.assign(key=[1]*df.shape[0])
        name_df = df.loc[:,["key","template"]]
        # 把需要的两列提取出来
        combs_df = name_df.merge(name_df,on="key",suffixes=("","_other")).loc[:,["template","template_other"]]
        # 去掉template和template_other相同的部分
        mask = (combs_df.iloc[:,0] == combs_df.iloc[:, 1])
        combs_df = combs_df.loc[~mask]

        fold_type_df = combs_df.apply(lambda d: d.apply(lambda x: key2value_dict[x]))
        
        # 比较fold_typd是否相同
        mask = (fold_type_df.iloc[:,0] == fold_type_df.iloc[:, 1])

        label_cons = tf.constant(mask,tf.int32)

        return label_cons
    
    def pair_between(self, lspec_df, rspec_df,key2value_dict):
        lspec_df = lspec_df.assign(key=[1]*lspec_df.shape[0]).loc[:,["key","template"]]
        rspec_df = rspec_df.assign(key=[1]*rspec_df.shape[0]).loc[:,["key","template"]]

        combs_df = lspec_df.merge(rspec_df,on="key",suffixes=("","_other")).loc[:,["template","template_other"]]
        # 去掉template和template_other相同的部分
        mask = (combs_df.iloc[:,0] == combs_df.iloc[:, 1])
        combs_df = combs_df.loc[~mask]

        fold_type_df = combs_df.apply(lambda d: d.apply(lambda x: key2value_dict[x]))
        
        # 比较fold_typd是否相同
        mask = (fold_type_df.iloc[:,0] == fold_type_df.iloc[:, 1])

        label_cons = tf.constant(mask,tf.int32)

        return label_cons        

    
    # 这里的X, y都是tensor类型
    # Eval中第一列表示label
    # Eval是tensor类型，但要用于训练过程中的验证集需要处理成[(X,y)]这样的格式
    # group和eGroup是ndarray类型，eGroup在用于训练的时候要特殊处理
    def fit(self,model, X, y, eX,ey,group, eGroup, early_stopping_rounds=100,verbose=10):
        X_ary = X.numpy()
        y_ary = y.numpy()
        eX_ary = eX.numpy()
        ey_ary = ey.numpy()
        eval_set = [(eX_ary,ey_ary)]
        eval_group = [eGroup]
        trained = model.fit(X_ary,y_ary,group=group,eval_set=eval_set,eval_group=eval_group,
            eval_metric="ndcg",eval_at=[10,50],early_stopping_rounds=early_stopping_rounds,
            verbose=10)
        return trained
    
    
    # X为tensor类型
    def predict(self, model, X):
        raw_score = model.predict(X.numpy(),raw_score=True)
        return raw_score
    
    # rep表示的是所有向量的特征, spec_file表示序列对应矩阵的哪一行
    # cv_file则数据集中的序列分别属于哪个子集
    # rep是一个ndarray或者tensor类型，cv_file则对应一个csvFormat文件的路径
    # 由于是mini batch的训练，因此，训练集的label也要分批次生成

    # mini batch的训练不行，现在改为通过hhblits的检索结果限制训练集的大小
    # srch_df表示检索得到的pairwise结果包含query, template, prob, evalue,score这几列
    def cv_training(self,model, rep, vrep, spec_df, vspec_df, srch_df, vsrch_df, cv_df, batch_size=1194):
        # r, c = spec_df.shape
        key2value_dict_path = "dataset/SCOP/dict/bin/sf/key2value"
        with open(key2value_dict_path,"rb") as h:
            kv_dict = cPickle.load(h)
        

        def cv_experiment(cv_df):
            for fold_num in cv_df.columns:
                s_df = srch_df.loc[~srch_df.loc[:,"query"].isin(cv_df.loc[:,fold_num])]
                s_df = s_df.loc[~s_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
                pairwise_df = s_df.loc[:,["query","template"]]

                vs_df = vsrch_df.loc[~vsrch_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
                vpairwise_df = vs_df.loc[:,["query","template"]]

                yield (pairwise_df, vpairwise_df)

        
        tr_list = []
        m_idx = 0
        mask = np.array([False]*5)
        select = [0,1,2,3,4]
        mask[select] = True
        for pairwise_df,vpairwise_df in cv_experiment(cv_df):
            if not mask[m_idx]:
                continue
            # start = 0
            # 去掉pairwise中query和template相同的情况
            idx = pairwise_df.loc[:,"query"] == pairwise_df.loc[:,"template"]

            fold_type_df = pairwise_df.apply(lambda d: d.apply(lambda x: kv_dict[x]))
            label_ss = fold_type_df.loc[~idx,"query"] == fold_type_df.loc[~idx,"template"]
            
            vfold_type_df = vpairwise_df.apply(lambda d: d.apply(lambda x: kv_dict[x]))
            vlabel_ss = vfold_type_df.loc[:,"query"] == vfold_type_df.loc[:,"template"]
            

            # 将名称转为行下标
            pairwise_df = \
                pairwise_df.apply(
                    lambda x: \
                        pd.Series(spec_df.loc[x,"row num"].reset_index(drop=True),
                            name=x.name)
                )
            
            # 这里要注意的是因为是validation检索training，因此，要分别得到行下标
            def val_get_row_num(x):
                if x.name == "query":
                    return pd.Series(vspec_df.loc[x,"row num"].reset_index(drop=True),
                        name=x.name)
                else:
                    return pd.Series(spec_df.loc[x,"row num"].reset_index(drop=True),
                        name=x.name)

            vpairwise_df = vpairwise_df.apply(val_get_row_num)                

            # 去掉pairwise中query和template相同的情况
            idx = pairwise_df.loc[:,"query"] == pairwise_df.loc[:,"template"]

            sim = self.pairwise_sim(pairwise_df.loc[~idx],rep)
            vsim = self.pairwise_sim(vpairwise_df,rep,vrep)

            label = tf.constant(label_ss,tf.int64)
            vlabel = tf.constant(vlabel_ss,tf.int64)


            group = pairwise_df.loc[~idx].groupby("query").apply(len).to_numpy(int)
            vgroup = vpairwise_df.groupby("query").apply(len).to_numpy(int)

            # model = lgb.LGBMClassifier(max_depth=500,objective="binary",n_jobs=10)
            # random_state = np.random.seed()
            # model = lgb.LGBMRanker(num_leaves=21,max_depth=400,
            #     objective="lambdarank",reg_alpha=5.7,reg_lambda=5.9,
            #     random_state=random_state,n_jobs=20)
            # firstFlag = True

            print(sim.shape,label.shape,vsim.shape,vlabel.shape,group.sum(),vgroup.sum())
            trained = self.fit(model,sim,label,vsim,vlabel,group,vgroup)

            m_idx += 1
            tr_list.append(trained)
        return tr_list

    # 调用lightgbm
    # import lightgbm as lgb
    # 然后设定params，调用lgb.train
    # 具体使用方法参考官方教程
    # 根据cross validation将数据集进行拆分，制作成lightgbm.Dataset格式
    # params是dict类型，用于lgb.train的params参数
    def cv_lambdarank(self, rep, vrep, spec_df, vspec_df, srch_df, vsrch_df, cv_df, 
        params):
        # r, c = spec_df.shape
        key2value_dict_path = "dataset/SCOP/dict/bin/sf/key2value"
        with open(key2value_dict_path,"rb") as h:
            kv_dict = cPickle.load(h)
        

        def cv_experiment(cv_df):
            for fold_num in cv_df.columns:
                s_df = srch_df.loc[~srch_df.loc[:,"query"].isin(cv_df.loc[:,fold_num])]
                s_df = s_df.loc[~s_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
                pairwise_df = s_df.loc[:,["query","template"]]

                vs_df = vsrch_df.loc[~vsrch_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
                vpairwise_df = vs_df.loc[:,["query","template"]]

                yield (pairwise_df, vpairwise_df)

        
        tr_list = []
        m_idx = 0
        mask = np.array([False]*5)
        select = [0,1,2,3,4]
        mask[select] = True
        for pairwise_df,vpairwise_df in cv_experiment(cv_df):
            if not mask[m_idx]:
                continue
            # start = 0
            # 去掉pairwise中query和template相同的情况
            idx = pairwise_df.loc[:,"query"] == pairwise_df.loc[:,"template"]

            fold_type_df = pairwise_df.apply(lambda d: d.apply(lambda x: kv_dict[x]))
            label_ss = fold_type_df.loc[~idx,"query"] == fold_type_df.loc[~idx,"template"]
            
            vfold_type_df = vpairwise_df.apply(lambda d: d.apply(lambda x: kv_dict[x]))
            vlabel_ss = vfold_type_df.loc[:,"query"] == vfold_type_df.loc[:,"template"]

            # 将名称转为行下标
            pairwise_df = \
                pairwise_df.apply(
                    lambda x: \
                        pd.Series(spec_df.loc[x,"row num"].reset_index(drop=True),
                            name=x.name)
                )
            
            # 这里要注意的是因为是validation检索training，因此，要分别得到行下标
            def val_get_row_num(x):
                if x.name == "query":
                    return pd.Series(vspec_df.loc[x,"row num"].reset_index(drop=True),
                        name=x.name)
                else:
                    return pd.Series(spec_df.loc[x,"row num"].reset_index(drop=True),
                        name=x.name)

            vpairwise_df = vpairwise_df.apply(val_get_row_num)                

            # 去掉pairwise中query和template相同的情况
            idx = pairwise_df.loc[:,"query"] == pairwise_df.loc[:,"template"]

            sim = self.pairwise_sim(pairwise_df.loc[~idx],rep)
            vsim = self.pairwise_sim(vpairwise_df,rep,vrep)

            label = tf.constant(label_ss,tf.int64)
            vlabel = tf.constant(vlabel_ss,tf.int64)


            group = pairwise_df.loc[~idx].groupby("query").apply(len).to_numpy(int)
            vgroup = vpairwise_df.groupby("query").apply(len).to_numpy(int)        

            # 转为lightgbm.Dataset
            lgb_train = lgb.Dataset(
                sim.numpy(),label.numpy(),group=group,free_raw_data=False
            )
            lgb_val = lgb.Dataset(
                vsim.numpy(),vlabel.numpy(),group=vgroup,free_raw_data=False
            )

            trained = lgb.train(
                params,
                lgb_train,
                valid_sets=[lgb_val],
                early_stopping_rounds=100,
                verbose_eval=10
            )
            
            tr_list.append(trained)
            m_idx += 1
        return tr_list
    
    def cv_lambdarank_predict(self, rep, spec_df, srch_df, cv_df, save_model_dir):
        # 预测时不需要知晓label
        # key2value_dict_path = "dataset/SCOP/dict/bin/sf/key2value"
        # with open(key2value_dict_path,"rb") as h:
        #     kv_dict = cPickle.load(h)
                
        # 用于predict时，query选择属于fold num，template选择不是fold num
        # 这样就符合交叉验证了
        # 预测的时候一条序列一条序列来
        def cv_experiment(cv_df):
            for fold_num in cv_df.columns:
                for queryName in cv_df.loc[:, fold_num]:
                    if pd.isna(queryName):
                        continue
                    # s_df = spec_df.reset_index().assign(query=queryName).rename(columns={"index":"template"})
                    # s_df = s_df.loc[~s_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
                    # pairwise_df = s_df.loc[:,["query","template"]]

                    # 同样通过searched result过滤一下
                    s_df = srch_df.loc[srch_df.loc[:,"query"] == queryName]
                    s_df = s_df.loc[~s_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
                    pairwise_df = s_df.loc[:,["query","template"]]

                    yield fold_num, pairwise_df
                
        pred_list = []
        for fold_num, pairwise_df in cv_experiment(cv_df):
            # 因为时交叉验证，不存在query和template相同的情况
            # start = 0
            # # 去掉pairwise中query和template相同的情况
            # idx = pairwise_df.loc[:,"query"] == pairwise_df.loc[:,"template"]

            # 将名称转为行下标
            p_df = \
                pairwise_df.apply(
                    lambda x: \
                        pd.Series(spec_df.loc[x,"row num"].reset_index(drop=True),
                            name=x.name)
                )
        
            # 理由同上
            # # 去掉pairwise中query和template相同的情况
            # idx = pairwise_df.loc[:,"query"] == pairwise_df.loc[:,"template"]

            sim = self.pairwise_sim(p_df,rep)
            
            trained_path = os.path.join(save_model_dir,"cv_0{}".format(fold_num))
            model = lgb.Booster(model_file=trained_path)
            y_pred = model.predict(sim.numpy(),raw_score=True)

            pred_list.append(y_pred)
        return pred_list

# 尝试通过LightGBM的Dataset构造数据集，同时进行增量训练
# 因为构建Dataset，有个free_raw_data，可以保证在构造Dataset后，释放原始数据，这保证了内存占用不会太大
class Ranker(object):
    # 根据所有的检索结果生成pairwise
    # srched_list是所有srch_df组成的列表
    # srch_df表示检索结果，是一个包含query, template的dataframe
    # vsrch_df格式与srch_df一样表示的是validation数据集检索training samples的结果
    def gen_pariwise(self, cv_df, srch_df, vsrch_df):
        for fold_num in cv_df.columns:
            idx_df = srch_df.apply(lambda x: x.isin(cv_df.loc[:,fold_num]))
            pairwise_df = srch_df.loc[~(idx_df.loc[:,"query"] == idx_df.loc[:,"template"])]
            vpairwise_df = vsrch_df.loc[~vsrch_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
            yield pairwise_df, vpairwise_df

    def similarity(self, **kArgs):
        assert kArgs.get("rep") is not None
        ron = tf.constant(kArgs["rep"],tf.float32)

        if kArgs.get("queryName") is not None and\
            kArgs.get("template_ser") is not None:
            tidx = kArgs["template_ser"].to_numpy().reshape((-1,1))
        
            dis = ron[kArgs["queryName"]] - tf.gather_nd(ron, tidx)
        elif kArgs.get("pairwise_df") is not None:
            pairwise_df = kArgs["pairwise_df"]
            lidx = pairwise_df.loc[:, "query"].to_numpy().reshape((-1,1))
            ridx = pairwise_df.loc[:, "template"].to_numpy().reshape((-1,1))
            dis = tf.gather_nd(ron, lidx) - tf.gather_nd(ron, ridx)
        elif kArgs.get("vpairwise_df") is not None and\
            kArgs.get("vrep") is not None:

            # print(kArgs)
            lidx = kArgs["vpairwise_df"].loc[:,"query"].to_numpy().reshape((-1,1))
            ridx = kArgs["vpairwise_df"].loc[:,"template"].to_numpy().reshape((-1,1))
            vron = tf.constant(kArgs["vrep"],tf.float32)
            
            # print(tf.gather_nd(vron, lidx))
            # print(tf.gather_nd(ron, ridx))
            dis = tf.gather_nd(vron, lidx) - tf.gather_nd(ron, ridx)
        else:
            raise AttributeError("可支持的参数，queryName, template_ser和rep或vpairwise_df,rep和vrep")
        
        dis = tf.abs(dis)
        maximum = tf.linalg.norm(dis, ord=np.inf,axis=1,keepdims=True)
        log = tf.math.log(
            tf.linalg.norm(
                dis,ord=2,axis=1,keepdims=True
            )
        )
        log /= tf.math.log(3.)
        sim = tf.concat([dis, maximum, log],axis=1)
            
        return sim
    
    
    # cv_prepare(self, queryName,train_df,pairwise_df, vpairwise_df,spec_df, vspec_df, kv_dict, kv_df):
    # train_df是query name为index，"行下标"为列的datagrame，表示trainin samples在特征矩阵中的位置
    # pairwise_df是包含query template这两列的dataframe，表示的是training samples构成的pairwise
    # vpairwise_df与pairwise_df类似，表示的是validation数据集与training samples构成的pairwise
    # spec_df与train_df一样
    # vspec_df与train_df格式一样，但表示的是validation数据集中的序列在特征矩阵中的位置
    # kv_dict和kv_df表示的是序列对应的超家族
    # 分别是字典(query name为key, superfamily为value)，dataframe格式(query name为index，superfamily为列)
    def cv_prepare(self, **kArgs):
        result = {} # 用于返回结果的字典类型文件

        assert (kArgs.get("kv_dict") is not None and \
            kArgs.get("kv_df") is not None)

        kv_dict = kArgs["kv_dict"]
        kv_df = kArgs["kv_df"]

        assert (kArgs.get("spec_df") is not None and\
            kArgs.get("vspec_df") is not None)
        spec_df = kArgs["spec_df"]
        vspec_df = kArgs["vspec_df"]

        if kArgs.get("queryName") is not None and \
            kArgs.get("train_df") is not None:
            queryName = kArgs.get("queryName")
            train_df = kArgs["train_df"]
            q_sf = kv_dict[queryName]
            label = (train_df.join(kv_df).loc[:,"sf"] == q_sf).to_numpy(np.int64)
            queryName = spec_df.loc[queryName,"row num"]
            result["queryName"] = queryName
            result["label"] = label
        elif kArgs.get("pairwise_df") is not None:
            pairwise_df = kArgs["pairwise_df"]
            pairwise_df = pairwise_df.join(kv_df,on="query")
            pairwise_df = pairwise_df.join(kv_df,lsuffix="_query",rsuffix="_template",on="template")
            label = (pairwise_df.loc[:, "sf_query"] == pairwise_df.loc[:,"sf_template"]).to_numpy(np.int64)

            pairwise_df = pairwise_df.join(vspec_df,on="query")
            pairwise_df = pairwise_df.join(spec_df,lsuffix="_query",rsuffix="_template",on="template")
            # print(pd.isna(pairwise_df).any())
            pairwise_df = pairwise_df.loc[:,["row num_query","row num_template"]]
            pairwise_df = pairwise_df.rename(columns={"row num_query":"query","row num_template":"template"})           
            result["label"] = label
            result["pairwise_df"] = pairwise_df
        else:
            raise AssertionError("queryName和train_df或pairwise_df必须提供一个")

        assert kArgs.get("vpairwise_df") is not None

        vpairwise_df = kArgs["vpairwise_df"]
        vpairwise_df = vpairwise_df.join(kv_df,on="query")
        vpairwise_df = vpairwise_df.join(kv_df,lsuffix="_query",rsuffix="_template",on="template")
        # print(pd.isna(vpairwise_df).any())
        vlabel = (vpairwise_df.loc[:, "sf_query"] == vpairwise_df.loc[:,"sf_template"]).to_numpy(np.int64)

        vpairwise_df = vpairwise_df.join(vspec_df,on="query")
        vpairwise_df = vpairwise_df.join(spec_df,lsuffix="_query",rsuffix="_template",on="template")
        # print(pd.isna(vpairwise_df).any())
        vpairwise_df = vpairwise_df.loc[:,["row num_query","row num_template"]]
        vpairwise_df = vpairwise_df.rename(columns={"row num_query":"query","row num_template":"template"})
        # print(pd.isna(vpairwise_df).any())

        result["vlabel"] = vlabel
        result["vpairwise_df"] = vpairwise_df

        return result

    # X, y, eX, ey都是numpy.ndarray的类型
    def cv_training(self, X,y, eX, ey,params,group=None,egroup=None,init_model=None):
        lgb_train = lgb.Dataset(
            X,y,group=group,free_raw_data=True
        )
        lgb_val = lgb.Dataset(
            eX,ey,group=egroup,free_raw_data=True
        )

        print(params)
        trained = lgb.train(
            params,
            lgb_train,
            valid_sets=[lgb_val],
            init_model=init_model,
            early_stopping_rounds=100,
            verbose_eval=10
        )

        # return lgb_train, lgb_val # 用于测试

        return trained

    # cv_df表示cross_validation的划分情况，一列表示一折
    # 在kArgs中
    # 提供了template_df则根据数据集划分情况生成对应用于训练的template_df
    # 提供了vtemplate_df则表示这是pairwise类型的检索结果，同样根据数据集划分情况进行生成
    # 体统了query_df则表示这是测试集，根据数据划分情况进行生成，注意区分query生成方法与其他两种的不同之处
    def training_generator(self, cv_df, **kArgs):
        assert kArgs.get("template_df") is not None

        template_df = kArgs["template_df"]
        result = {}

        assert "template" in template_df.columns

        for fold_num in cv_df.columns:
            # 如果存在query
            # 就表示这是pairwise类型的检索结果
            # 否则就表示这是只有template(或者还有row num)的dataframe
            if "query" in template_df.columns:
                template_df = template_df.loc[~template_df.loc[:,"query"].isin(cv_df.loc[:,fold_num])]
            template_df = template_df.loc[~template_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
            result["template"] = template_df

            # validaiont数据集如果有query也不用处理，因为validatoin数据集和training数据集是不重叠的
            vtemplate_df = kArgs["vtemplate_df"]
            vtemplate_df = vtemplate_df.loc[~vtemplate_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])]
            result["vteamplte"] = vtemplate_df

            yield result
    
    def data_generator(self, representation_l, val_representation_l=None):
        valFlag = False
        for i, representation_path in enumerate(representation_l):
            if val_representation_l is not None:
                vrepresentation_path = val_representation_l[i]
                vrep = np.load(vrepresentation_path)
                valFlag = True
            rep = np.load(representation_path)

            if valFlag:
                yield rep, vrep
            else:
                yield (rep, None)
    
    def get_label_by_cv(self,cvr, kv_df):
        template_df = cvr["template"]
        vtemplate_df = cvr["vtemplate"]

        if "query" in template_df.columns:
            template_df = template_df.join(kv_df.set_index("template"), on="query")
            template_df = template_df.join(kv_df.set_index("template"), on="template", lsuffix="_query",rsuffix="_template")
            label_ss = template_df.loc[:, "sf_query"] == template_df.loc[:, "sf_template"]

            vtemplate_df = vtemplate_df.join(kv_df.set_index("template"), on="query")
            vtemplate_df = vtemplate_df.join(kv_df.set_index("template"), on="template", lsuffix="_query",rsuffix="_template")
            vlabel_ss = vtemplate_df.loc[:, "sf_query"] == vtemplate_df.loc[:, "sf_template"]

            y = label_ss.to_numpy(dtype=np.int64)
            ey = vlabel_ss.to_numpy(dtype=np.int64)
        else:
            template_df = template_df.join(kv_df, on="template")
            vtemplate_df = vtemplate_df.join(kv_df, on="template")

            m_df = pd.concat([template_df.loc[:,"sf"], vtemplate_df.loc[:,"sf"]]).rename("sf")
            m_df = m_df.reset_index()
            num_class = len(m_df.groupby("sf"))
            m_df = m_df.assign(new_label=list(range(num_class)))
            template_df = template_df.join(m_df.set_index("sf"),on="sf")
            vtemplate_df = vtemplate_df.join(m_df.set_index("sf"),on="sf")

            y = template_df.loc[:, "new_label"].to_numpy(np.int64)
            ey = template_df.loc[:, "new_label"].to_numpy(np.int64)
        return y, ey

    
    def get_data_by_cv(self, cvr, rep, vrep,spec_df, vspec_df):
        template_df = cvr["template"]
        vtemplate_df = cvr["vtemplate"]

        if "query" in template_df.columns:
            # 需要调用sim
            # 因为这表示这是pairwise类型的检索结果
            template_df = template_df.join(spec_df,on="query")
            template_df = template_df.join(spec_df,on="template", lsuffix="_query",rsuffix="_template")
            template_df = \
                template_df.loc[:["row num_query","row num_template"]].rename(columns={"row num_query":"query","row num_template": "template"})
            X = self.similarity(pairwise_df = template_df, rep=rep)

            vtemplate_df = vtemplate_df.join(spec_df,on="query")
            vtemplate_df = vtemplate_df.join(spec_df,on="template", lsuffix="_query",rsuffix="_template")
            vtemplate_df = \
                vtemplate_df.loc[:["row num_query","row num_template"]].rename(columns={"row num_query":"query","row num_template": "template"})
            eX = self.similarity(vpairwise_df = vtemplate_df, vrep=vrep)            
        else:
            template_df = template_df.join(spec_df,on="template")
            vtemplate_df = vtemplate_df.join(vspec_df,on="template")

            idx = template_df.loc[:,"row num"].to_numpy(np.int64)
            vidx = template_df.loc[:, "row num"].to_numpy(np.int64)

            X = rep[idx]
            eX = vrep[vidx]
        
        return X, eX

        

    def cv_model_training(self, params, init_model, cv_df, spec_df, vspec_df, kv_df, 
        representaion_l, val_representaion_l, save_l, **kArgs):

        if kArgs.get("template_df") is not None:
            template_df = kArgs["template_df"]

        else:
            template_df = spec_df.reset_index().rename(columns={"index":"template"})
            
        
        if kArgs.get("vtemplate_df") is not None:
            vtemplate_df = kArgs["vtemplate_df"]
        else:
            vtemplate_df = vspec_df.reset_index().rename(columns={"index":"template"})

        for cvr in self.training_generator(cv_df=cv_df, template_df=template_df,vtemplate_df=vtemplate_df):
            y, ey = self.get_label_by_cv(cvr, kv_df)
            for rep, vrep in self.data_generator(representaion_l, val_representaion_l):
                X, eX = self.get_data_by_cv(cvr, rep, vrep, spec_df, vspec_df)


    def cli_cv_training(self,params,cv_df, spec_df, vspec_df, kv_df, conf_path,
        representaion_l, val_representaion_l,save_l):
        flag = False
        tunning = True
        valid_df = vspec_df.join(kv_df.set_index("template"))
        vlabel = valid_df.loc[:,"label"].to_numpy(np.int64)
        # vgroup = [valid_df.shape[0]]
        template_ss = pd.Series(spec_df.index,name="template")

        def gen_training(cv_df):
            count = 5
            for fold_num in cv_df.columns:
                if flag and count == 0:
                    break
                train_df = spec_df.loc[template_ss.loc[~template_ss.isin(cv_df.loc[:,fold_num])]]

                yield fold_num, train_df
                if flag:
                    count -= 1

        sel_ary = np.array([[False]] * 4)
        idx = [0]
        sel_ary[idx] = True
        # client = Client(processes=False)             # create local cluster
        for fold_num, train_df in gen_training(cv_df):
            train_df = train_df.join(kv_df.set_index("template"))
            # 重新给superfamily编号
            # 但是这里一开始忘记给vlabel重新编号了，因此，初步评估的结果并不准确需要重新评估
            k_df = train_df.groupby("sf").apply(lambda x: x.iloc[0,0]).reset_index()
            # print(k_df)
            k_df = k_df.reset_index().loc[:,["sf","index"]]
            # print(k_df)
            train_df = train_df.join(k_df.set_index("sf"),on="sf")
            train_df = train_df.rename(columns={"index":"new label"})
            classes = [x for x in train_df.groupby("new label").groups.keys()]
            label = train_df.loc[:,"new label"].to_numpy(np.int64)
            
            num_class = train_df.loc[:, "new label"].max()+1
            # params["num_class"] = num_class # SVM训练不需要知道num_class
            # label = label_binarize(label, classes=list(range(num_class))) # 用于方便计算roc_auc
            # vlabel = label_binarize(vlabel, classes=list(range(num_class))) # 用于方便计算roc_auc
            # group = [train_df.shape[0]]
            for i, representaion_path in enumerate(representaion_l):
                if not sel_ary[i]:
                    continue
                # print(vpairwise_df)
                val_representaion_path = val_representaion_l[i]
                rep = np.load(representaion_path)
                vrep = np.load(val_representaion_path)
                # print(vpairwise_df)
                idx = train_df.loc[:,"row num"].to_numpy(int)
                rep = rep[idx]

                # lgb_train, lgb_test = run_ltr.cv_training(sim,label,group,
                #     vsim,vlabel,egroup,params) # 用于测试
                
                printStart = "printf '数据集处理完毕\n'"
                sp.call(printStart,shell=True)
                saveDir = save_l[i]
                if tunning:
                    #用于parameters tunning
                    time_suffix = time.strftime("%Y%m%d-%H:%M:%S",time.localtime())
                    savePath = os.path.join(
                        saveDir,"cv_0{}_{}".format(fold_num,time_suffix)
                    )
                else:
                    # 输出到optimal路径下，就表示参数都确定了
                    savePath = os.path.join(
                        saveDir,"cv_0{}".format(fold_num)
                    )
                # training_dataset_path = os.path.join(os.path.dirname(representaion_path),"multiclass.train")
                # validation_dataset_path = os.path.join(os.path.dirname(representaion_path),"multiclass.valid")

                # params["data"] = training_dataset_path
                # params["valid_data"] = validation_dataset_path
                # params["output_model"] = savePath
                
                # data = lgb.Dataset(rep,label,free_raw_data=False)
                # data.construct()
                # valid_data = lgb.Dataset(vrep,vlabel,free_raw_data=False)
                # valid_data.construct()

                # if os.path.exists(training_dataset_path):
                #     rm = "rm {}".format(training_dataset_path)
                #     sp.call(rm, shell=True)

                # data.save_binary(training_dataset_path)

                # if os.path.exists(validation_dataset_path):
                #     rm = "rm {}".format(validation_dataset_path)
                #     sp.call(rm, shell=True)

                # valid_data.save_binary(validation_dataset_path)

                # 归一化
                norms = self.row_norms(rep,[1,2,np.inf])
                norm = np.mean(norms, axis=1,keepdims=True)
                rep = norm * rep

                norms = self.row_norms(vrep,[1,2,np.inf])
                norm = np.mean(norms, axis=1,keepdims=True)
                vrep = norm * vrep
                printStart = "printf '归一化完毕\n'"
                sp.call(printStart,shell=True)

                # # 保存为对应的格式用于LightGBM的训练
                # rep = np.hstack([label.reshape((-1,1)),rep])
                # vrep = np.hstack([vlabel.reshape((-1,1)),vrep])

                # print(rep.shape)
                # print(vrep.shape)
                # forceFlag = True
                # if forceFlag or not os.path.exists(training_dataset_path):
                #     self.construct_dataset(rep,training_dataset_path)
                # if forceFlag or not os.path.exists(validation_dataset_path):
                #     self.construct_dataset(vrep,validation_dataset_path)
                
                # # 再把params保存到conf_path中
                # self.save_conf(params, conf_path)

                # printStart = "printf '开始训练\n'"
                # sp.call(printStart,shell=True)

                # # 命令行调用
                # training = "lightgbm config={}".format(conf_path)
                # sp.call(training,shell=True)
                
                # # 调用sklearn的GradientBoostingClassifier
                # clf = GradientBoostingClassifier(**params)
                # printStart = "printf '开始训练\n'"
                # sp.call(printStart,shell=True)
                # with joblib.parallel_backend('dask'):
                #     clf.fit(rep, label)
                
                # printStart = "printf '利用验证集评估\n'"
                # sp.call(printStart,shell=True)
                # score = clf.decision_function(vrep)
                
                # auc_mu = self.auc_mu(score, vlabel, classes)
                # print("auc: {}".format(auc_mu))

                # with open(savePath, "wb") as h:
                #     cPickle.dump(clf, h)

                # 调用svm
                self.sklearn_svm(params, rep, label, vrep,vlabel, classes,savePath)
    
    def sklearn_svm(self,params, X,y, eX, ey, classes, savePath):
        clf = LinearSVC(**params)
        printStart = "printf '开始训练\n'"
        sp.call(printStart,shell=True)
        clf.fit(X, y)

        printStart = "printf '保存模型到{}\n'".format(savePath)
        sp.call(printStart,shell=True)
        with open(savePath, "wb") as h:
            cPickle.dump(clf, h)

        printStart = "printf '开始评估\n'"
        sp.call(printStart,shell=True)
        score = clf.decision_function(eX)

        auc_mu = self.auc_mu(score, ey, classes)
        printStart = "printf {}".format(auc_mu)
        sp.call(printStart,shell=True)
        

    def save_conf(self, params, conf_path):
        with open(conf_path, "w") as h:
            for key in params:
                line = "{} = {}\n".format(key,params[key])
                h.write(line)
    
    def construct_dataset(self, data, dataset_path):
        r, c = data.shape
        format = "\t".join(["%.4f"]*(c-1))
        with open(dataset_path,"w") as h:
            for i in range(r):
                x = data[i]
                label = int(x[0])
                feature = tuple(x[1:])
                f_str = format % feature
                line = "{}\t{}\n".format(label,f_str)
                h.write(line)
            print(label)
    
    def row_norms(self, mat, ords):
        norms = np.hstack(
            [
                np.linalg.norm(mat,ord,axis=1,keepdims=True)\
                    for ord in ords
            ]
        )
        return norms
    
    def auc_mu(self, y_pred, y_true, classes):
        y_true = label_binarize(y_true, classes=classes)
        fpr = dict()
        tpr = dict()
        roc_auc = []
        for i in range(len(classes)):
            fpr[i], tpr[i], _ = roc_curve(y_true[:, i], y_pred[:, i])
            if not np.isnan(tpr[i]).any():
                try:
                    roc_auc.append(auc(fpr[i], tpr[i]))
                except:
                    if 1 in tpr[i]:
                        roc_auc.append(1.)
                    else:
                        roc_auc.append(0.)
        auc_score = np.array(roc_auc).mean()
        return auc_score

