from Bio import SeqIO as sio
import pandas as pd
import numpy as np
from six.moves import cPickle
import os
import subprocess as sp
import re


class Assembly(object):
    def __init__(self,featureDir=None):
        if featureDir is None:
            self.featureDir = featureDir
        self.level2no = {"fold":2,"sf":3,"fam":4}
    
    def fastaFileSize(self, fastaFilePath):
        with open(fastaFilePath,"r") as h:
            count = 0
            for record in sio.parse(h,"fasta"):
                count += 1
        return count

    def fasta2dict(self, fastaFilePath,dictFilePath=None,level="sf",binary=False):
        if dictFilePath is None:
            assert self.featureDir is not None
            dictFilePath = self.featureDir
        level_no = self.level2no[level]
        key2value = {}
        value2key = {}
        for record in sio.parse(fastaFilePath,"fasta"):
            des = record.description
            scop_id, fa_id = des.split()[:2]
            sf_id = ".".join(fa_id.split(".")[:level_no])
            key2value[scop_id] = sf_id
            
            if value2key.get(sf_id) is None:
                value2key[sf_id] = []
            value2key[sf_id].append(scop_id)
        
        target = {
            "key2value": key2value,
            "value2key": value2key
        }
        
        for k in target:
            if binary:
                cPickle.dump(
                    target[k],
                    open(
                        os.path.join(dictFilePath,k),
                        "wb"
                    )
                )
            else:
                with open(os.path.join(dictFilePath,k),"w") as h:
                    for k1 in target[k]:
                        line = "{} {}\n".format(k1,target[k][k1])
                        h.write(line)
    
    #对于oriDictPath中的文件，supportDictPath是过滤后的，如果oriDictPath在level下，能在
    #supprotDictPath对应的文件中找到对应的部分，则保留，反之则删除
    #一般这样用，oriDictPath提供了famliy2seq的文件，supportDictPath提供了supperfamily2seq的文件
    #但supperfamliy2seq是已删除了部分supperfamily的文件，那么根据这个文件，把family2seq文件进行过滤
    #因此level选择sf,表示superfamily
    #最终过滤后的文件会输出到upDictPath中
    #所有文件默认通过cPickle读取
    def changeLevel2Dict(self, oriDictPath, supportDictPath,upDictPath,level="sf"):
        level_no = self.level2no[level]
        up_d = {}
        with open(oriDictPath,"rb") as h, \
            open(supportDictPath, "rb") as h1, \
            open(upDictPath,"wb") as h2:
            ori_d = cPickle.load(h)
            sup_d = cPickle.load(h1)
            for k in ori_d:
                lk = ".".join(k.split(".")[:level_no])
                if sup_d.get(lk) is not None:
                    if up_d.get(k) is None:
                        up_d[k] = []
                    up_d[k] += ori_d[k]
            cPickle.dump(up_d,h2)
    
    def split_n_subset(self, n, datasetDictPath,result_dir,binary=True):        
        with open(datasetDictPath,"rb") as h:
            dataset_d = cPickle.load(h)
        
        subsetSize_l = [0 for _ in range(n)]
        subsetNo = ["sub_{}".format(x) for x in range(n)]
        seq2no = {}
        no2seq = {}
        subset2keyDict = {}
        
        key2size = sorted([(k, len(dataset_d[k])) for k in dataset_d],
            key=lambda x:x[1],reverse=True)
        
        def countErrors(L):
            comb = self.combination(L,2)
            sum = 0
            for p in comb:
                sum += (abs(p[0]-p[1]))
            return sum
        
        #计算不同子集之间的差的和
        for key, size in key2size:
            #一开始分配给第一个
            min_i = 0
            min_err = -1
            for i in range(n):
                subsetSize_l[i] += size
                current_err = countErrors(subsetSize_l)
                if min_err == -1 or current_err < min_err:
                    min_i = i
                    min_err = current_err
                subsetSize_l[i] -= size

            subsetSize_l[min_i] += size
            if subset2keyDict.get(subsetNo[min_i]) is None:
                subset2keyDict[subsetNo[min_i]] = []
            subset2keyDict[subsetNo[min_i]].append(key)
        
        for subset in subset2keyDict:
            if no2seq.get(subset) is None:
                no2seq[subset] = []
            for x in subset2keyDict[subset]:
                for y in dataset_d[x]:
                    seq2no[y] = subset
                no2seq[subset] += dataset_d[x]
        
        if binary:
            dict_path = os.path.join(result_dir,"dict","binary","seq2no")
            with open(dict_path,"wb") as h:
                cPickle.dump(seq2no,h)
        else:
            dict_path = os.path.join(result_dir,"dict","text","seq2no")
            with open(dict_path,"w") as h:
                for k in seq2no:
                    line = "{} {}\n".format(k,seq2no[k])
                    h.write(line)
        
        # 使列表保证一样的长度
        m_size = max([len(no2seq[x]) for x in no2seq])
        none_l = [None] * m_size
        for k in no2seq:
            size = len(no2seq[k])
            if size < m_size:
                none_l[:size] = no2seq[k]
                no2seq[k] = none_l
            none_l = [None] * m_size
        sn_df = pd.DataFrame(no2seq)
        
        csv_path = os.path.join(result_dir,"csv","no2seq.csv")
        sn_df.to_csv(
            open(csv_path,"w",newline=''),
            index=False
        )
    
    
    #by是一个dict的pickle文件
    #Key表示seq名称，value表示该seq所属的子集名称
    def split_file_by(self,fastaFilePath,by,resultDir):
        no2path = {}
        path2record = {}
        seq2no = cPickle.load(open(by,"rb"))
        with open(fastaFilePath,"r") as h:
            for record in sio.parse(h,"fasta"):
                name = record.name
                if seq2no.get(name) is None:
                    continue
                no = seq2no[name]
                if no2path.get(no) is None:
                    subset_path = os.path.join(resultDir,no)
                    # if os.path.exists(subset_path):
                    #     os.remove(subset_path)
                    no2path[no] = subset_path
                    path2record[subset_path] = []
                subset_path = no2path[no]
                path2record[subset_path].append(record)
        
        for path in path2record:
            with open(path, "w") as h:
                sio.write(path2record[path],h,"fasta")
    
    def use_cdhit(self,fastaFilePath,outputFilePath,sequence_identity=0.95,word_size=5,threads=10):
        # 这里要进行判断当sequence_identity低于0.4时要调用另一个工具psi-cd-hit.pl
        # sp.call("cd-hit --help",shell=True) #简单测试
        if sequence_identity >= 0.4:
            cmd = "cd-hit -i {} -o {} -c {} -n {} -T {}".format(
                fastaFilePath,outputFilePath,
                sequence_identity,word_size, threads
            )
        else:
            cmd = "psi-cd-hit.pl -i {} -o {} -c {} -blp {} -prog psiblast".format(
                fastaFilePath,outputFilePath,sequence_identity,threads
            )
        sp.call(cmd,shell=True)
    
    def use_makeblastdb(self,inputFilePath,outputFilePath):
        cmd = "makeblastdb -in {} -dbtype 'prot' -out {}".format(inputFilePath,outputFilePath)
        sp.call(cmd,shell=True)
    
    def use_splitfasta(self,fastaFilePath,sequencesDir):
        cmd = "splitfasta.pl {}".format(fastaFilePath)
        currentDir = os.getcwd()
        os.chdir(sequencesDir)
        sp.call(cmd,shell=True)
        os.chdir(currentDir)
    
    def splitFasta2seqs(self, fastaFilePath,sequencesDir):
        from Bio.SeqIO import parse, write
        from Bio.Alphabet import generic_protein

        with open(fastaFilePath,"r") as h:
            for record in parse(h,"fasta"):
                name = record.name
                seqFilePath = os.path.join(sequencesDir,"{}.seq".format(name))
                with open(seqFilePath,"w") as h1:
                    write(record,h1,"fasta")
    
    def use_psiblast(self, queryFilePath, dbpath, outputPath,evalue=1e-4):
        cmd = "psiblast -query {} -db {} -out {} -evalue {}".format(
            queryFilePath,dbpath,outputPath,evalue
        )
        sp.call(cmd,shell=True)

    #把psi文件转为dict
    def psi2set(self, psiFilePath,templateSet=set()):
        #d4xkma_ c.1.15.0
        # qm = re.compile(r"Query\s*=")
        m0 = r"Sequences producing significant alignments"
        m = r"(?P<name>[a-zA-Z0-9_\.]+)\s+(?P<fam>[a-z]\.[0-9]+\.[0-9]+\.[0-9]+)"
        m1 = r"^>"
        m2 = r"No hits found"
        start_flag = False
        # not_empty_flag = False
        
        tplSet = templateSet
        # queryName = None
        with open(psiFilePath, "r") as h:
            for l in h:
                m0_r = re.search(m0,l)
                m1_r = re.search(m1,l)
                m2_r = re.search(m2,l)
                # if qm.search(l) is not None:
                #     m_r = re.search(m, l)
                #     if m_r is not None:
                #         queryName = m_r.group("name")
                if m0_r is not None:
                    start_flag = True
                if start_flag:
                    m_r = re.search(m, l)
                    if m_r is not None:
                        name = m_r.group("name")
                        tplSet.add(name)
                        # not_empty_flag = True
                if m1_r is not None or m2_r is not None:
                    break

        # if not_empty_flag and (queryName is not None):
        #     tplSet.add(queryName)

        return tplSet
    
    def mmseqs_cluster(self, fastaFilePath,resultDir,sequence_identity=0.95,evalue=1e-4):
        fastaFileName = os.path.basename(fastaFilePath)
        m = r"\."
        m_r = re.search(m, fastaFileName)
        if m_r is not None:
            fastaFileName = fastaFileName[:m_r.start()]
        # fastaDBPath = os.path.join(os.path.dirname(os.path.dirname(fastaFilePath)),"db",
        #     "{}_db".format(fastaFileName))
        # create_fasta_db_cmd = "mmseqs createdb {} {}".format(fastaFilePath,fastaDBPath)
        # sp.call(create_fasta_db_cmd,shell=True)
        
        tmpDir = os.path.join(os.path.dirname(fastaFilePath),
            "tmp")
        if os.path.exists(tmpDir):
            os.rmdir(tmpDir)
        os.mkdir(tmpDir)

        #mmseqs cluster DB clu tmp --min-seq-id 0.95 -e 0.0001
        # clusterFilePath = os.path.join(os.path.dirname(os.path.dirname(fastaFilePath)),
        #     "clu","{}_clu".format(fastaFileName))
        clusterFilePath = os.path.join(resultDir,fastaFileName)
        # cluster_db_cmd = "mmseqs cluster {} {} {} --min-seq-id {} -e {}".format(
        #     fastaDBPath,clusterFilePath,tmpDir,sequence_identity,evalue
        # )
        cluster_cmd = "mmseqs easy-cluster {} {} {} --min-seq-id {} -e {}".format(
            fastaFilePath,clusterFilePath,tmpDir,sequence_identity,evalue
        )
        sp.call(cluster_cmd,shell=True)

        # #mmseqs createseqfiledb DB clu clu_seq
        # cluster_seqFilePath = os.path.join(os.path.dirname(os.path.dirname(fastaFilePath)),"clu",
        #     "{}_clu_seq".format(fastaFileName))
        # createseqfile_cmd = "mmseqs createseqfiledb {} {} {}".format(
        #     fastaDBPath,clusterFilePath,cluster_seqFilePath
        # )
        # sp.call(createseqfile_cmd,shell=True)

        # # mmseqs result2flat DB DB clu_seq clu_seq.fasta
        # upFastaFilePath = os.path.join(resultDir,fastaFileName)
        # result2flat_cmd = "mmseqs result2flat {} {} {} {}".format(
        #     fastaDBPath,fastaDBPath,cluster_seqFilePath,upFastaFilePath
        # )
        # sp.call(result2flat_cmd,shell=True)

        # # mmseqs createtsv DB DB clu clu.tsv
        # tsvPath = os.path.join(resultDir,"{}_clu.tsv".format(fastaFileName))
        # createtsv_cmd = "mmseqs createtsv {} {} {} {}".format(
        #     fastaDBPath,fastaDBPath,clusterFilePath,tsvPath
        # )
        # sp.call(createtsv_cmd,shell=True)

        #删除tmp
        rmtmp_cmd = "rm -rf {}".format(tmpDir)
        sp.call(rmtmp_cmd,shell=True)

    
    #根据by的内容剔除数据集里的序列
    def filterDatasetBy(self, datasetPath, by, outputPath):
        record_list = []
        with open(datasetPath,"r") as h:
            for record in sio.parse(h,"fasta"):
                name = record.name
                if name in by:
                    continue
                record_list.append(record)
        with open(outputPath,"w") as h:
            sio.write(record_list,h,"fasta")
    
    #mmseq cluster的结果中有多余的record
    def filterFastaRecord(self, fastaFilePath,resultPath):
        with open(fastaFilePath,"r") as h,\
            open(resultPath,"w") as h1:
            for record in sio.parse(h,"fasta"):
                if record.seq == "":
                    continue
                sio.write(record,h1,"fasta")
    
    # 返回一个dataframe格式的文件
    # 包含template, sf, label三列
    def superfamilyName2label(self, kv_dict):
        kv_df = pd.DataFrame(kv_dict.items(),columns=["template","sf"])
        # kv_df
        k_df = kv_df.groupby("sf").apply(lambda x: x.loc[0,"template"]).reset_index()
        k_df = k_df.reset_index().loc[:,["sf","index"]]
        kv_df = kv_df.join(k_df.set_index("sf"),on="sf")
        return kv_df.rename(columns={"index":"label"})
    
    # 根据cross validation的内容，重新编号
    def superfamilyName2newLabel(self,kv_df, cv_df):
        def gen_cv(cv_df):
            for fold_num in cv_df.columns:
                idx = ~kv_df.loc[:,"template"].isin(cv_df.loc[:,fold_num])
                yield idx
        train_l = []
        for idx in gen_cv(cv_df):
            train_df = kv_df.loc[idx]
            # print(train_df)
            # print(train_df.groupby("sf").apply(lambda x: x.iloc[0,0]))
            k_df = train_df.groupby("sf").apply(lambda x: x.iloc[0,0]).reset_index()
            # print(k_df)
            k_df = k_df.reset_index().loc[:,["sf","index"]]
            # print(k_df)
            train_df = train_df.join(k_df.set_index("sf"),on="sf")
            # print(train_df)
            train_l.append(train_df.rename(columns={"index":"new label"}))
        return train_l
    
    # 读取fasPath
    # 将该文件转为template, fam, sf, fold, fam_label, sf_label, fold_labe这样的dataframe
    def fas2info_csvFormat(self, fasPath):
        from Bio.SeqIO import parse, write
        from Bio.Alphabet import generic_protein
        import pandas as pd
        def get_info(record):
            des = record.description
            name, fam = des.split()[:2]
            pointer = re.compile(r"\.")
            L = pointer.split(fam)
            sf = ".".join(L[:3])
            fold = ".".join(L[:2])
            return [name, fam,sf,fold]
        
        with open(fasPath,"r") as h:
            info_list = list(
                map(get_info,parse(h,"fasta"))
            )
        
        info_df = pd.DataFrame(info_list,columns=["template","fam","sf","fold"])
        
        # 编号
        for k in info_df.columns[1:]:
            ig = info_df.groupby(k)
            g_df = pd.DataFrame(list(ig.groups.keys()),columns=[k])
            col_name = "{}_label".format(k)
            g_df = g_df.reset_index().rename(columns={"index":col_name})
            info_df = info_df.join(g_df.set_index(k),on=k)
        
        return info_df
            
    
    def combination(self, L, k):
        size = len(L)
        if k == 0:
            return [[]]
        if size == 0:
            return []
        comb = []
        
        x = L[0]
        res = L[1:]
        part = [[x] + y for y in self.combination(res,k-1)]
        comb = part + self.combination(res,k)
        return comb



